#!/usr/bin/env python

# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  in_toto_sign.py

<Author>
  Sachit Malik <i.sachitmalik@gmail.com>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  June 13, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides command line interface to sign in-toto link or layout metadata
  or to verify its signatures.

"""
import argparse
import logging
import sys
from getpass import getpass

from securesystemslib._gpg import functions as gpg_interface

from in_toto import __version__, exceptions
from in_toto.common_args import (
    GPG_HOME_ARGS,
    GPG_HOME_KWARGS,
    QUIET_ARGS,
    QUIET_KWARGS,
    VERBOSE_ARGS,
    VERBOSE_KWARGS,
    sort_action_groups,
    title_case_action_groups,
)
from in_toto.formats import _check_hex
from in_toto.models._signer import (
    GPGSigner,
    load_crypto_signer_from_pkcs8_file,
    load_public_key_from_file,
)
from in_toto.models.link import FILENAME_FORMAT
from in_toto.models.metadata import Metadata

# Command line interfaces should use in_toto base logger (c.f. in_toto.log)
LOG = logging.getLogger("in_toto")


def _sign_and_dump_metadata(metadata, args):
    """
    <Purpose>
      Internal method to sign link or layout metadata and dump it to disk.

    <Arguments>
      metadata:
              Metadata object (contains Link or Layout object)
      args:
              see argparser

    <Exceptions>
      SystemExit(0) if signing is successful
      SystemExit(2) if any exception occurs

    """

    try:
        if not args.append:
            metadata.signatures.clear()

        signature = None
        # If the cli tool was called with `--gpg [KEYID ...]` `args.gpg` is
        # a list (not None) and we will try to sign with gpg.
        # If `--gpg-home` was not set, args.gpg_home is None and the signer tries
        # to use the default gpg keyring.
        if args.gpg is not None:
            # If `--gpg` was passed without argument we sign with the default key
            # Excluded so that coverage does not vary in different test environments
            if len(args.gpg) == 0:  # pragma: no cover
                signature = metadata.create_signature(
                    GPGSigner(None, homedir=args.gpg_home)
                )

            # Otherwise we sign with each passed keyid
            for keyid in args.gpg:
                _check_hex(keyid)
                signature = metadata.create_signature(
                    GPGSigner(keyid, homedir=args.gpg_home)
                )

        # Alternatively we iterate over passed private key paths `--key KEYPATH
        # ...` load the corresponding key from disk and sign with it
        elif args.key is not None:  # pragma: no branch
            password = None
            if args.prompt:
                password = getpass()
                password = password.encode()

            for path in args.key:
                signer = load_crypto_signer_from_pkcs8_file(path, password)
                signature = metadata.create_signature(signer)

        payload = metadata.get_payload()
        _type = payload.type_

        # If `--output` was specified we store the signed link or layout metadata
        # to that location no matter what
        if args.output:
            out_path = args.output

        # Otherwise, in case of links, we build the filename using the link/step
        # name and the keyid of the created signature (there is only one for links)
        elif _type == "link":
            keyid = signature.keyid
            out_path = FILENAME_FORMAT.format(
                step_name=payload.name, keyid=keyid
            )

        # In case of layouts we just override the input file.
        elif _type == "layout":  # pragma: no branch
            out_path = args.file

        LOG.info("Dumping %s to '%s'...", _type, out_path)

        metadata.dump(out_path)
        sys.exit(0)

    except Exception as e:  # pylint: disable=broad-exception-caught
        LOG.error("The following error occurred while signing: %s", e)
        sys.exit(2)


def _verify_metadata(metadata, args):
    """
    <Purpose>
      Internal method to verify link or layout signatures.

    <Arguments>
      metadata:
              Metadata object (contains Link or Layout object)
      args:
              see argparser

    <Exceptions>
      SystemExit(0) if verification passes
      SystemExit(1) if verification fails
      SystemExit(2) if any exception occurs

    """
    try:
        # Load pubkeys from disk ....
        if args.key is not None:
            pub_key_dict = {}
            for path in args.key:
                key = load_public_key_from_file(path)
                pub_key_dict[key["keyid"]] = key

        # ... or from gpg keyring
        elif args.gpg is not None:  # pragma: no branch
            pub_key_dict = gpg_interface.export_pubkeys(args.gpg, args.gpg_home)

        for keyid, verification_key in pub_key_dict.items():
            metadata.verify_signature(verification_key)
            LOG.info("Signature verification passed for keyid '%s'", keyid)

        sys.exit(0)

    except exceptions.SignatureVerificationError as e:
        LOG.error("Signature verification failed: %s", e)
        sys.exit(1)

    except Exception as e:  # pylint: disable=broad-exception-caught
        LOG.error(
            "The following error occurred while verifying signatures: %s", e
        )
        sys.exit(2)


def _load_metadata(file_path):
    """
    <Purpose>
      Loads Metadata (link or layout metadata) file from disk

    <Arguments>
      file_path:
              path to link or layout metadata file

    <Exceptions>
      SystemExit(2) if any exception occurs

    <Returns>
      in-toto Metadata object (contains Link or Layout object)

    """
    try:
        return Metadata.load(file_path)

    except Exception as e:  # pylint: disable=broad-exception-caught
        LOG.error(
            "The following error occurred while loading the file '%s': %s",
            file_path,
            e,
        )
        sys.exit(2)


def create_parser():
    """Create and return configured ArgumentParser instance."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
in-toto-sign provides a command line interface to sign in-toto link or layout
metadata or verify their signatures, with options to:

  * replace (default) or add signatures:

    + layout metadata can be signed by multiple keys at once,
    + link metadata can only be signed by one key at a time.

  * write signed metadata to a specified path. If no output path is specified,

    + layout metadata is written to the path of the input file,
    + link metadata is written to '<name>.<keyid prefix>.link'.

  * verify signatures

in-toto-sign is useful to re-sign metadata (e.g. when changing keys), or to
sign unsigned links (e.g. generated with 'in-toto-mock'). For layouts, it is
useful to append signatures in case threshold signing of layouts is necessary.

It returns a non-zero value on failure and zero otherwise.""",
    )

    parser.epilog = """EXAMPLE USAGE

Sign 'unsigned.layout' with two keys and write it to 'root.layout'.

  {prog} -f unsigned.layout -k priv_key1 priv_key2 -o root.layout


Replace signature in link file and write to default filename, i.e.
'package.<priv_key keyid prefix>.link'.

  {prog} -f package.2f89b927.link -k priv_key


Verify layout signed with 3 keys.

  {prog} -f root.layout -k pub_key0 pub_key1 pub_key2 --verify


Sign layout with default gpg key in default gpg keyring.

  {prog} -f root.layout --gpg


Verify layout with a gpg key identified by keyid '...439F3C2'.

  {prog} -f root.layout --verify \\
      --gpg 3BF8135765A07E21BD12BF89A5627F6BF439F3C2

""".format(
        prog=parser.prog
    )

    named_args = parser.add_argument_group("required named arguments")

    named_args.add_argument(
        "-f",
        "--file",
        type=str,
        required=True,
        metavar="<path>",
        help=("path to link or layout file to be signed or verified."),
    )

    parser.add_argument(
        "-k",
        "--key",
        nargs="+",
        metavar="<path>",
        help=(
            "paths to key files, used to sign the passed link or layout metadata"
            " or to verify its signatures. Verification keys are expected in"
            " PEM/subjectPublicKeyInfo and signing keys in PEM/PKCS8 format."
            " Pass '--prompt' to enter a signing key decryption password."
        ),
    )

    parser.add_argument(
        "-p",
        "--prompt",
        action="store_true",
        help="prompt for signing key decryption password",
    )

    parser.add_argument(
        "-g",
        "--gpg",
        nargs="*",
        metavar="<id>",
        help=(
            "GPG keyids used to sign the passed link or layout metadata or to verify"
            " its signatures. If passed without keyids, the default GPG key is"
            " used."
        ),
    )

    parser.add_argument(*GPG_HOME_ARGS, **GPG_HOME_KWARGS)

    # Only when signing
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        metavar="<path>",
        help=(
            "path to location where the metadata file is stored after signing. If"
            " not passed, layout metadata is written to the path of the input file"
            " and link metadata is written to '<name>.<keyid prefix>.link'"
        ),
    )

    # Only when signing
    parser.add_argument(
        "-a",
        "--append",
        action="store_true",
        help=(
            "add signatures rather than replacing existing signatures. This option"
            " is only availabe for layout metdata."
        ),
    )

    parser.add_argument(
        "--verify",
        action="store_true",
        help="verify signatures of passed link or layout metadata using the"
        " public keys passed via '--key' and/or '--gpg' options.",
    )

    verbosity_args = parser.add_mutually_exclusive_group(required=False)
    verbosity_args.add_argument(*VERBOSE_ARGS, **VERBOSE_KWARGS)
    verbosity_args.add_argument(*QUIET_ARGS, **QUIET_KWARGS)

    parser.add_argument(
        "--version",
        action="version",
        version="{} {}".format(parser.prog, __version__),
    )

    title_case_action_groups(parser)
    sort_action_groups(parser)

    return parser


def main():
    """Parse arguments, load link or layout metadata file and either sign
    metadata file or verify its signatures."""

    parser = create_parser()
    args = parser.parse_args()

    LOG.setLevelVerboseOrQuiet(args.verbose, args.quiet)

    # Additional argparse sanitization
    # NOTE: This tool is starting to have many inter-dependent argument
    # restrictions. Maybe we should make it less sophisticated at some point.
    if args.verify and (args.append or args.output):
        parser.print_help()
        parser.error(
            "conflicting arguments: don't specify any of"
            " 'append' or 'output' when verifying signatures"
        )

    # Regular signing and GPG signing are mutually exclusive
    if (args.key is None) == (args.gpg is None):
        parser.print_help()
        parser.error(
            "wrong arguments: specify either '--key PATH [PATH ...]'"
            " or '--gpg [KEYID [KEYID ...]]'"
        )

    # For gpg verification we must specify a keyid (no default key is loaded)
    if args.verify and args.gpg is not None and len(args.gpg) < 1:
        parser.print_help()
        parser.error(
            "missing arguments: specify at least one keyid for GPG"
            " signature verification ('--gpg KEYID ...')"
        )

    metadata = _load_metadata(args.file)

    # Specific command line argument restrictions if we deal with links
    payload = metadata.get_payload()
    if payload.type_ == "link":
        # Above we check that it's either `--key ...` or `--gpg ...`
        # Here we check that it is not more than one in each case when dealing
        # with links
        link_error_message = (
            "link metadata is associated with a"
            " single functionary and is usually namespaced accordingly:"
            " '<name>.<keyid prefix>.link'."
        )

        if (args.key is not None and len(args.key) > 1) or (
            args.gpg is not None and len(args.gpg) > 1
        ):
            parser.print_help()
            parser.error(
                "too many arguments: {} Hence signing Link metadata"
                " with multiple keys is not allowed.".format(link_error_message)
            )

        if args.append:
            parser.print_help()
            parser.error(
                "wrong arguments: {}. Hence adding signatures to"
                " existing signatures on link metadata is not allowed.".format(
                    link_error_message
                )
            )

    if args.verify:
        _verify_metadata(metadata, args)

    else:
        _sign_and_dump_metadata(metadata, args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  in_toto_mock.py

<Author>
  Shikher Verma <root@shikherverma.com>

<Started>
  June 12, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides a command line interface for runlib.in_toto_mock.

<Return Codes>
  2 if an exception occurred during argument parsing
  1 if an exception occurred
  0 if no exception occurred

"""
import argparse
import logging
import sys

import in_toto.runlib
from in_toto import __version__
from in_toto.common_args import (
    DSSE_ARGS,
    DSSE_KWARGS,
    sort_action_groups,
    title_case_action_groups,
)

# Command line interfaces should use in_toto base logger (c.f. in_toto.log)
LOG = logging.getLogger("in_toto")


def create_parser():
    """Parse arguments and call in_toto_mock."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
in-toto-mock is a variant of 'in-toto-run' that can be used to create unsigned
link metadata, using defaults for many of the 'in-toto-run' arguments.
in-toto-mock verbosely executes the passed command, records all files in the
current working directory as materials and products, and generates a link file
under '<name>.link'.

This is useful for trying out how to generate a link without the need for a
key, or knowledge about all 'in-toto-run' arguments. It can also be used to
quickly generate link metadata, inspect it and sign it retroactively.

""",
    )

    parser.usage = "%(prog)s [-h] --name <name> -- <command> [args]"

    parser.epilog = """EXAMPLE USAGE

Generate unsigned link metadata 'foo.link' for the activity of creating file
'bar', inspect it, and sign it with 'mykey'

  # Generate unsigned link
  {prog} --name foo -- touch bar
  # Inspect and/or update unsigned link metadata
  vi foo.link
  # Sign the link, attesting to its validity, and write it to
  # 'foo.<mykey keyid prefix>.link'.
  in-toto-sign -k mykey -f foo.link

""".format(
        prog=parser.prog
    )

    named_args = parser.add_argument_group("required named arguments")

    # FIXME: Do we limit the allowed characters for the name?
    named_args.add_argument(
        "-n",
        "--name",
        type=str,
        required=True,
        metavar="<name>",
        help=(
            "name for the resulting link metadata file, which is written to"
            " '<name>.link'. It is also used to associate the link with a step"
            " defined in an in-toto layout."
        ),
    )

    parser.add_argument(*DSSE_ARGS, **DSSE_KWARGS)

    # FIXME: This is not yet ideal.
    # What should we do with tokens like > or ; ?
    parser.add_argument(
        "link_cmd",
        nargs="+",
        metavar="<command>",
        help=(
            "command to be executed. It is separated from named and optional"
            " arguments by a double dash '--'."
        ),
    )

    parser.add_argument(
        "--version",
        action="version",
        version="{} {}".format(parser.prog, __version__),
    )

    title_case_action_groups(parser)
    sort_action_groups(parser)

    return parser


def main():
    """Parse arguments and call in_toto_mock."""
    parser = create_parser()
    args = parser.parse_args()

    # in-toto-mock should not be used to secure the supply chain but only to try
    # out in-toto-run with max. user feedback, hence we set a verbose log level
    LOG.setLevel(logging.INFO)

    try:
        in_toto.runlib.in_toto_mock(args.name, args.link_cmd, args.use_dsse)

    except Exception as e:  # pylint: disable=broad-exception-caught
        LOG.error("(in-toto-mock) %s: %s", type(e).__name__, e)
        sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""CLI to check local artifacts.
"""
import argparse
import logging
import sys

from in_toto.common_args import (
    LSTRIP_PATHS_ARGS,
    LSTRIP_PATHS_KWARGS,
    VERBOSE_ARGS,
    VERBOSE_KWARGS,
    sort_action_groups,
    title_case_action_groups,
)
from in_toto.models.metadata import Metadata
from in_toto.runlib import in_toto_match_products as match_products

LOG = logging.getLogger(__name__)


def create_parser():
    """Create parser."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Check if local artifacts match products in passed link.",
    )
    required = parser.add_argument_group("required named arguments")
    required.add_argument(
        "-l",
        "--link",
        type=str,
        required=True,
        metavar="<path>",
        help="path to link metadata file.",
    )
    parser.add_argument(
        "-p",
        "--paths",
        type=str,
        required=False,
        nargs="+",
        metavar="<path>",
        help="file or directory paths to local artifacts. Default is CWD.",
    )

    parser.add_argument(
        "-e",
        "--exclude",
        required=False,
        metavar="<pattern>",
        nargs="+",
        help="gitignore-style patterns to exclude artifacts from matching.",
    )

    parser.add_argument(*LSTRIP_PATHS_ARGS, **LSTRIP_PATHS_KWARGS)
    parser.add_argument(*VERBOSE_ARGS, **VERBOSE_KWARGS)

    title_case_action_groups(parser)
    sort_action_groups(parser)

    return parser


def main():
    """CLI."""
    parser = create_parser()
    args = parser.parse_args()

    metadata = Metadata.load(args.link)
    only_products, not_in_products, differ = match_products(
        metadata.signed,
        paths=args.paths,
        exclude_patterns=args.exclude,
        lstrip_paths=args.lstrip_paths,
    )
    # raise Exception("test")
    if only_products or not_in_products or differ:
        if args.verbose:
            print("Local artifacts don't match products in passed link.")
            for name in only_products:
                print(f"Only in products: {name}")

            for name in not_in_products:
                print(f"Not in products: {name}")

            for name in differ:
                print(f"Hashes differ: {name}")

        sys.exit(1)

    if args.verbose:
        print("Local artifacts match products in passed link.")
    sys.exit(0)


if __name__ == "__main__":
    main()



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  runlib.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  June 27, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides a wrapper for any command of the software supply chain.

  The wrapper performs the following tasks which are implemented in this
  library.

    - Record state of material (files the command is executed on)
    - Execute command
      - Capture stdout/stderr/return value of the executed command
    - Record state of product (files after the command was executed)
    - Return Metadata containing a Link object which can be can be signed
      and stored to disk
"""
import glob
import io
import logging
import os
import subprocess  # nosec
import sys
import tempfile
import time
from collections import defaultdict

import securesystemslib._gpg
import securesystemslib.exceptions
import securesystemslib.formats
import securesystemslib.hash
from securesystemslib.signer import Signer

import in_toto.exceptions
import in_toto.settings
from in_toto.formats import _check_hex, _check_str, _check_str_list
from in_toto.models._signer import GPGSigner
from in_toto.models.link import (
    FILENAME_FORMAT,
    FILENAME_FORMAT_SHORT,
    UNFINISHED_FILENAME_FORMAT,
    UNFINISHED_FILENAME_FORMAT_GLOB,
)
from in_toto.models.metadata import Envelope, Metablock, Metadata
from in_toto.resolver import (
    RESOLVER_FOR_URI_SCHEME,
    DirectoryResolver,
    FileResolver,
    OSTreeResolver,
    Resolver,
)

# Inherits from in_toto base logger (c.f. in_toto.log)
LOG = logging.getLogger(__name__)


def record_artifacts_as_dict(
    artifacts,
    exclude_patterns=None,
    base_path=None,
    follow_symlink_dirs=False,
    normalize_line_endings=False,
    lstrip_paths=None,
):
    """
    <Purpose>
      Hashes each file in the passed path list. If the path list contains
      paths to directories the directory tree(s) are traversed.

      The files a link command is executed on are called materials.
      The files that result form a link command execution are called
      products.

      Paths are normalized for matching and storing by left stripping "./"

      NOTE on exclude patterns:
        - Uses PathSpec to compile gitignore-style patterns, making use of the
          GitWildMatchPattern class (registered as 'gitwildmatch')

        - Patterns are checked for match against the full path relative to each
          path passed in the artifacts list

        - If a directory is excluded, all its files and subdirectories are also
          excluded

        - How it differs from .gitignore
              - No need to escape #
              - No ignoring of trailing spaces
              - No general negation with exclamation mark !
              - No special treatment of slash /
              - No special treatment of consecutive asterisks **

        - Exclude patterns are likely to become command line arguments or part of
          a config file.

    <Arguments>
      artifacts:
              A list of file or directory paths used as materials or products for
              the link command.

      exclude_patterns: (optional)
              Artifacts matched by the pattern are excluded from the result.
              Exclude patterns can be passed as argument or specified via
              ARTIFACT_EXCLUDE_PATTERNS setting (see `in_toto.settings`).
              If passed, patterns specified via settings are overriden.

      base_path: (optional)
              Change to base_path and record artifacts relative from there.
              If not passed, current working directory is used as base_path.
              NOTE: The base_path part of the recorded artifact is not included
              in the returned paths.

      follow_symlink_dirs: (optional)
              Follow symlinked dirs if the linked dir exists (default is False).
              The recorded path contains the symlink name, not the resolved name.
              NOTE: This parameter toggles following linked directories only,
              linked files are always recorded, independently of this parameter.
              NOTE: Beware of infinite recursions that can occur if a symlink
              points to a parent directory or itself.

      normalize_line_endings: (optional)
              If True, replaces windows and mac line endings with unix line
              endings before hashing the content of the passed files, for
              cross-platform support.

      lstrip_paths: (optional)
              If a prefix path is passed, the prefix is left stripped from
              the path of every artifact that contains the prefix.

    <Exceptions>
      OSError: cannot change to base path directory.
      ValueError: arguments are malformed.

    <Side Effects>
      Calls functions to generate cryptographic hashes.

    <Returns>
      A dictionary with file paths as keys and the files' hashes as values.

    """
    artifact_hashes = {}

    if not artifacts:
        return artifact_hashes

    if not base_path:
        base_path = in_toto.settings.ARTIFACT_BASE_PATH

    if not exclude_patterns:
        exclude_patterns = in_toto.settings.ARTIFACT_EXCLUDE_PATTERNS

    # Configure resolver with resolver specific arguments
    # FIXME: This should happen closer to the user boundary, where
    # resolver-specific config arguments are passed and global state is managed.
    RESOLVER_FOR_URI_SCHEME[FileResolver.SCHEME] = FileResolver(
        exclude_patterns,
        base_path,
        follow_symlink_dirs,
        normalize_line_endings,
        lstrip_paths,
    )

    # Configure resolver for OSTree
    RESOLVER_FOR_URI_SCHEME[OSTreeResolver.SCHEME] = OSTreeResolver(base_path)

    # Configure resolver for hashing directories as a single entry
    RESOLVER_FOR_URI_SCHEME[DirectoryResolver.SCHEME] = DirectoryResolver(
        exclude_patterns=exclude_patterns,
        follow_symlink_dirs=follow_symlink_dirs,
        normalize_line_endings=normalize_line_endings,
        lstrip_paths=lstrip_paths,
    )

    # Aggregate artifacts per resolver
    resolver_for_uris = defaultdict(list)
    for artifact in artifacts:
        resolver = Resolver.for_uri(artifact)
        resolver_for_uris[resolver].append(artifact)

    # Hash artifacts in a batch per resolver
    # FIXME: The behavior may change if we hash each artifact individually,
    # because the left-prefix duplicate check in FileResolver only works for the
    # artifacts hashed in one batch.
    for resolver, uris in resolver_for_uris.items():
        artifact_hashes.update(resolver.hash_artifacts(uris))

    # Clear resolvers to not preserve global state change beyond this function.
    # FIXME: This also clears resolver registered elsewhere. For now we
    # assume that we only modify RESOLVER_FOR_URI_SCHEME in this function.
    RESOLVER_FOR_URI_SCHEME.clear()

    return artifact_hashes


def _subprocess_run_duplicate_streams(cmd, timeout):
    """Helper to run subprocess and both print and capture standards streams.

    Caveat:
    * Might behave unexpectedly with interactive commands.
    * Might not duplicate output in real time, if the command buffers it (see
      e.g. `print("foo")` vs. `print("foo", flush=True)`).
    * Possible race condition on Windows when removing temporary files.

    """
    # Use temporary files as targets for child process standard stream redirects
    # They seem to work better (i.e. do not hang) than pipes, when using
    # interactive commands like `vi`.
    stdout_fd, stdout_name = tempfile.mkstemp()
    stderr_fd, stderr_name = tempfile.mkstemp()
    try:
        with io.open(  # pylint: disable=unspecified-encoding
            stdout_name, "r"
        ) as stdout_reader, os.fdopen(  # pylint: disable=unspecified-encoding
            stdout_fd, "w"
        ) as stdout_writer, io.open(  # pylint: disable=unspecified-encoding
            stderr_name, "r"
        ) as stderr_reader, os.fdopen(
            stderr_fd, "w"
        ) as stderr_writer:
            # Store stream results in mutable dict to update it inside nested helper
            streams = {"out": "", "err": ""}

            def _duplicate_streams():
                """Helper to read from child process standard streams, write their
                contents to parent process standard streams, and build up return values
                for outer function.
                """
                # Read until EOF but at most `io.DEFAULT_BUFFER_SIZE` bytes per call.
                # Reading and writing in reasonably sized chunks prevents us from
                # subverting a timeout, due to being busy for too long or indefinitely.
                stdout_part = stdout_reader.read(io.DEFAULT_BUFFER_SIZE)
                stderr_part = stderr_reader.read(io.DEFAULT_BUFFER_SIZE)
                sys.stdout.write(stdout_part)
                sys.stderr.write(stderr_part)
                sys.stdout.flush()
                sys.stderr.flush()
                streams["out"] += stdout_part
                streams["err"] += stderr_part

            # Start child process, writing its standard streams to temporary files
            proc = subprocess.Popen(  # pylint: disable=consider-using-with  # nosec
                cmd,
                stdout=stdout_writer,
                stderr=stderr_writer,
                universal_newlines=True,
            )
            proc_start_time = time.time()

            # Duplicate streams until the process exits (or times out)
            while proc.poll() is None:
                # Time out as Python's `subprocess.run` would do it
                if (
                    timeout is not None
                    and time.time() > proc_start_time + timeout
                ):
                    proc.kill()
                    proc.wait()
                    raise subprocess.TimeoutExpired(cmd, timeout)

                _duplicate_streams()

            # Read/write once more to grab everything that the process wrote between
            # our last read in the loop and exiting, i.e. breaking the loop.
            _duplicate_streams()

    finally:
        # The work is done or was interrupted, the temp files can be removed
        # FIXME: retry failed file removal once to maybe work around #547
        for name in (stdout_name, stderr_name):
            try:
                os.remove(name)
            except PermissionError:  # pragma: no cover
                time.sleep(0.01)
                os.remove(name)

    # Return process exit code and captured streams
    return proc.poll(), streams["out"], streams["err"]


def execute_link(link_cmd_args, record_streams, timeout):
    """
    <Purpose>
      Executes the passed command plus arguments in a subprocess and returns
      the return value of the executed command. If the specified standard output
      and standard error of the command are recorded and also returned to the
      caller.

    <Arguments>
      link_cmd_args:
              A list where the first element is a command and the remaining
              elements are arguments passed to that command.
      record_streams:
              A bool that specifies whether to redirect standard output and
              and standard error to a temporary file which is returned to the
              caller (True) or not (False).

    <Exceptions>
      OSError:
              The given command is not present or non-executable

      subprocess.TimeoutExpired:
              The execution of the given command times.

    <Side Effects>
      Executes passed command in a subprocess and redirects stdout and stderr
      if specified.

    <Returns>
      - A dictionary containing standard output and standard error of the
        executed command, called by-products.
        Note: If record_streams is False, the dict values are empty strings.
      - The return value of the executed command.
    """
    if record_streams:
        return_code, stdout_str, stderr_str = _subprocess_run_duplicate_streams(
            link_cmd_args, timeout=timeout
        )

    else:
        process = subprocess.run(
            link_cmd_args,
            check=False,  # nosec
            timeout=timeout,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        stdout_str = stderr_str = ""
        return_code = process.returncode

    return {
        "stdout": stdout_str,
        "stderr": stderr_str,
        "return-value": return_code,
    }


def in_toto_mock(name, link_cmd_args, use_dsse=False):
    """
    <Purpose>
      in_toto_run with defaults
       - Records materials and products in current directory
       - Does not sign resulting link file
       - Stores resulting link file under "<name>.link"

    <Arguments>
      name:
              A unique name to relate mock link metadata with a step or
              inspection defined in the layout.
      link_cmd_args:
              A list where the first element is a command and the remaining
              elements are arguments passed to that command.
      use_dsse (optional):
              A boolean indicating if DSSE should be used to generate metadata.

    <Exceptions>
      None.

    <Side Effects>
      Writes newly created link metadata file to disk using the filename scheme
      from link.FILENAME_FORMAT_SHORT

    <Returns>
      Newly created Metadata object containing a Link object

    """
    link_metadata = in_toto_run(
        name,
        ["."],
        ["."],
        link_cmd_args,
        record_streams=True,
        use_dsse=use_dsse,
    )

    filename = FILENAME_FORMAT_SHORT.format(step_name=name)
    LOG.info("Storing unsigned link metadata to '%s'...", filename)
    link_metadata.dump(filename)
    return link_metadata


def _check_signer(signer):
    if not isinstance(signer, Signer):
        raise ValueError("signer must be a Signer instance")


def _require_signing_arg(signer, gpg_keyid, gpg_use_default):
    if not any([signer, gpg_keyid, gpg_use_default]):
        raise ValueError(
            "Pass either a signer, a gpg keyid or set gpg_use_default to True!"
        )


def in_toto_run(
    name,
    material_list,
    product_list,
    link_cmd_args,
    record_streams=False,
    gpg_keyid=None,
    gpg_use_default=False,
    gpg_home=None,
    exclude_patterns=None,
    base_path=None,
    compact_json=False,
    record_environment=False,
    normalize_line_endings=False,
    lstrip_paths=None,
    metadata_directory=None,
    use_dsse=False,
    timeout=in_toto.settings.LINK_CMD_EXEC_TIMEOUT,
    signer=None,
):
    """Performs a supply chain step or inspection generating link metadata.

  Executes link_cmd_args, recording paths and hashes of files before and after
  command execution (aka. artifacts) in a link metadata file. The metadata is
  signed with the passed signer, a gpg key identified by its ID, or
  the default gpg key. If multiple key arguments are passed, only one key is
  used in above order of precedence. The resulting link file is written to
  ``STEP-NAME.KEYID-PREFIX.link``. If no key argument is passed the link
  metadata is neither signed nor written to disk.

  Arguments:
    name: A unique name to associate link metadata with a step or inspection.

    material_list: A list of artifact paths to be recorded before command
        execution. Directories are traversed recursively.

    product_list: A list of artifact paths to be recorded after command
        execution. Directories are traversed recursively.

    link_cmd_args: A list where the first element is a command and the
        remaining elements are arguments passed to that command.

    record_streams (optional): A boolean indicating if standard output and
        standard error of the link command should be recorded in the link
        metadata in addition to being displayed while the command is executed.

    gpg_keyid (optional): A keyid used to identify a local gpg key used to sign
        the resulting link metadata.

    gpg_use_default (optional): A boolean indicating if the default gpg key
        should be used to sign the resulting link metadata.

    gpg_home (optional): A path to the gpg home directory. If not set the
        default gpg home directory is used.

    exclude_patterns (optional): A list of filename patterns to exclude certain
        files from being recorded as artifacts. See Config docs for details.

    base_path (optional): A path relative to which artifacts are recorded.
        Default is the current working directory.

    compact_json (optional): A boolean indicating if the resulting link
        metadata should be written in the most compact JSON representation.

    record_environment (optional): A boolean indicating if information about
        the environment should be added in the resulting link metadata.

    normalize_line_endings (optional): A boolean indicating if line endings of
        artifacts should be normalized before hashing for cross-platform
        support.

    lstrip_paths (optional): A list of path prefixes used to left-strip
        artifact paths before storing them in the resulting link metadata.

    metadata_directory (optional): A directory path to write the resulting link
        metadata file to. Default destination is the current working directory.

    use_dsse (optional): A boolean indicating if DSSE should be used to
        generate metadata.

    timeout (optional): An integer indicating the max timeout in seconds
        for this command. Default is 10 seconds.

    signer (optional): A securesystemslib Signer instance used to
        sign the resulting link metadata.

  Raises:
    securesystemslib.exceptions.FormatError: Passed arguments are malformed.

    OSError: Cannot change to base path directory.

    securesystemslib.exceptions.StorageError: Cannot hash artifacts.

    PrefixError: Left-stripping artifact paths results in non-unique dict keys.

    subprocess.TimeoutExpired: Link command times out.

    IOError, FileNotFoundError, NotADirectoryError, PermissionError:
        Cannot write link metadata.

    securesystemslib.exceptions.CryptoError, \
            securesystemslib.exceptions.UnsupportedAlgorithmError:
        Signing errors.

    ValueError, OSError, securesystemslib._gpg.exceptions.CommandError, \
            securesystemslib._gpg.exceptions.KeyNotFoundError:
        gpg signing errors.

  Side Effects:
    Reads artifact files from disk.
    Runs link command in subprocess.
    Calls system gpg in a subprocess, if a gpg key argument is passed.
    Writes link metadata file to disk, if any key argument is passed.

  Returns:
    A Metadata object that contains the resulting link object.

  """
    # pylint: disable=too-many-branches, too-many-locals, too-many-statements

    LOG.info("Running '%s'...", name)

    # Check key formats to fail early
    if signer:
        _check_signer(signer)

    if gpg_keyid:
        _check_hex(gpg_keyid)

    if exclude_patterns:
        _check_str_list(exclude_patterns)

    if base_path:
        _check_str(base_path)

    if metadata_directory:
        _check_str(metadata_directory)

    if material_list:
        LOG.info("Recording materials '%s'...", ", ".join(material_list))

    materials_dict = record_artifacts_as_dict(
        material_list,
        exclude_patterns=exclude_patterns,
        base_path=base_path,
        follow_symlink_dirs=True,
        normalize_line_endings=normalize_line_endings,
        lstrip_paths=lstrip_paths,
    )

    if link_cmd_args:
        _check_str_list(link_cmd_args)
        LOG.info("Running command '%s'...", " ".join(link_cmd_args))
        byproducts = execute_link(link_cmd_args, record_streams, timeout)
    else:
        byproducts = {}

    if product_list:
        _check_str_list(product_list)
        LOG.info("Recording products '%s'...", ", ".join(product_list))

    products_dict = record_artifacts_as_dict(
        product_list,
        exclude_patterns=exclude_patterns,
        base_path=base_path,
        follow_symlink_dirs=True,
        normalize_line_endings=normalize_line_endings,
        lstrip_paths=lstrip_paths,
    )

    LOG.info("Creating link metadata...")
    environment = {}
    if record_environment:
        environment["workdir"] = os.getcwd().replace("\\", "/")

    link = in_toto.models.link.Link(
        name=name,
        materials=materials_dict,
        products=products_dict,
        command=link_cmd_args,
        byproducts=byproducts,
        environment=environment,
    )

    if use_dsse:
        LOG.info("Generating link metadata using DSSE...")
        link_metadata = Envelope.from_signable(link)
    else:
        LOG.info("Generating link metadata using Metablock...")
        link_metadata = Metablock(signed=link, compact_json=compact_json)

    if signer:
        LOG.info("Signing link metadata using passed signer...")

    elif gpg_keyid:
        LOG.info("Signing link metadata using passed GPG keyid...")
        signer = GPGSigner(keyid=gpg_keyid, homedir=gpg_home)

    elif gpg_use_default:
        LOG.info("Signing link metadata using default GPG key ...")
        signer = GPGSigner(keyid=None, homedir=gpg_home)

    # We need the signature's keyid to write the link to keyid infix'ed filename
    if signer:
        signature = link_metadata.create_signature(signer)
        signing_keyid = signature.keyid

        filename = FILENAME_FORMAT.format(step_name=name, keyid=signing_keyid)

        if metadata_directory is not None:
            filename = os.path.join(metadata_directory, filename)

        LOG.info("Storing link metadata to '%s'...", filename)
        link_metadata.dump(filename)

    return link_metadata


def in_toto_record_start(
    step_name,
    material_list,
    gpg_keyid=None,
    gpg_use_default=False,
    gpg_home=None,
    exclude_patterns=None,
    base_path=None,
    record_environment=False,
    normalize_line_endings=False,
    lstrip_paths=None,
    use_dsse=False,
    signer=None,
):
    """Generates preliminary link metadata.

  Records paths and hashes of materials in a preliminary link metadata file. The
  metadata is signed with the passed signer, a gpg key identified
  by its ID, or the default gpg key. If multiple key arguments are passed, only
  one key is used in above order of precedence. At least one key argument must
  be passed. The resulting link file is written to
  ``.STEP-NAME.KEYID-PREFIX.link-unfinished``.

  Use this function together with in_toto_record_stop as an alternative to
  in_toto_run, in order to provide evidence for supply chain steps that cannot
  be carried out by a single command.

  Arguments:
    step_name: A unique name to associate link metadata with a step.

    material_list: A list of artifact paths to be recorded as materials.
        Directories are traversed recursively.

    gpg_keyid (optional): A keyid used to identify a local gpg key used to sign
        the resulting link metadata.

    gpg_use_default (optional): A boolean indicating if the default gpg key
        should be used to sign the resulting link metadata.

    gpg_home (optional): A path to the gpg home directory. If not set the
        default gpg home directory is used.

    exclude_patterns (optional): A list of filename patterns to exclude certain
        files from being recorded as artifacts. See Config docs for details.

    base_path (optional): A path relative to which artifacts are recorded.
        Default is the current working directory.

    record_environment (optional): A boolean indicating if information about
        the environment should be added in the resulting link metadata.

    normalize_line_endings (optional): A boolean indicating if line endings of
        artifacts should be normalized before hashing for cross-platform
        support.

    lstrip_paths (optional): A list of path prefixes used to left-strip
        artifact paths before storing them in the resulting link metadata.

    use_dsse (optional): A boolean indicating if DSSE should be used to
        generate metadata.

    signer (optional): A securesystemslib Signer instance used to
        sign the resulting link metadata.

  Raises:
    securesystemslib.exceptions.FormatError: Passed arguments are malformed.

    ValueError: None of signer, gpg_keyid or gpg_use_default=True is
        passed.

    securesystemslib.exceptions.StorageError: Cannot hash artifacts.

    PrefixError: Left-stripping artifact paths results in non-unique dict keys.

    subprocess.TimeoutExpired: Link command times out.

    IOError, PermissionError:
        Cannot write link metadata.

    securesystemslib.exceptions.CryptoError, \
            securesystemslib.exceptions.UnsupportedAlgorithmError:
        Signing errors.

    ValueError, OSError, securesystemslib._gpg.exceptions.CommandError, \
            securesystemslib._gpg.exceptions.KeyNotFoundError:
        gpg signing errors.

  Side Effects:
    Reads artifact files from disk.
    Calls system gpg in a subprocess, if a gpg key argument is passed.
    Writes preliminary link metadata file to disk.

  """
    # pylint: disable=too-many-locals,too-many-branches

    LOG.info("Start recording '%s'...", step_name)

    # Fail if there is no signing key arg at all
    _require_signing_arg(signer, gpg_keyid, gpg_use_default)

    # Check key formats to fail early
    if signer:
        _check_signer(signer)

    if gpg_keyid:
        _check_str(gpg_keyid)

    if exclude_patterns:
        _check_str_list(exclude_patterns)

    if base_path:
        _check_str(base_path)

    if material_list:
        LOG.info("Recording materials '%s'...", ", ".join(material_list))

    materials_dict = record_artifacts_as_dict(
        material_list,
        exclude_patterns=exclude_patterns,
        base_path=base_path,
        follow_symlink_dirs=True,
        normalize_line_endings=normalize_line_endings,
        lstrip_paths=lstrip_paths,
    )

    LOG.info("Creating preliminary link metadata...")
    environment = {}
    if record_environment:
        environment["workdir"] = os.getcwd().replace("\\", "/")

    link = in_toto.models.link.Link(
        name=step_name,
        materials=materials_dict,
        products={},
        command=[],
        byproducts={},
        environment=environment,
    )

    if use_dsse:
        LOG.info("Generating link metadata using DSSE...")
        link_metadata = Envelope.from_signable(link)
    else:
        LOG.info("Generating link metadata using Metablock...")
        link_metadata = Metablock(signed=link)

    if signer:
        LOG.info("Signing link metadata using passed signer...")

    elif gpg_keyid:
        LOG.info("Signing link metadata using passed GPG keyid...")
        signer = GPGSigner(keyid=gpg_keyid, homedir=gpg_home)

    else:  # (gpg_use_default)
        LOG.info("Signing link metadata using default GPG key ...")
        signer = GPGSigner(keyid=None, homedir=gpg_home)

    signature = link_metadata.create_signature(signer)
    # We need the signature's keyid to write the link to keyid infix'ed filename
    signing_keyid = signature.keyid

    unfinished_fn = UNFINISHED_FILENAME_FORMAT.format(
        step_name=step_name, keyid=signing_keyid
    )

    LOG.info("Storing preliminary link metadata to '%s'...", unfinished_fn)
    link_metadata.dump(unfinished_fn)


def in_toto_record_stop(
    step_name,
    product_list,
    gpg_keyid=None,
    gpg_use_default=False,
    gpg_home=None,
    exclude_patterns=None,
    base_path=None,
    normalize_line_endings=False,
    lstrip_paths=None,
    metadata_directory=None,
    command=None,
    byproducts=None,
    environment=None,
    signer=None,
):
    """Finalizes preliminary link metadata generated with in_toto_record_start.

  Loads preliminary link metadata file, verifies its signature, and records
  paths and hashes as products, thus finalizing the link metadata. The metadata
  is signed with the passed signer, a gpg key identified by its ID,
  or the default gpg key. If multiple key arguments are passed, only one key is
  used in above order of precedence. At least one key argument must be passed
  and it must be the same as the one used to sign the preliminary link metadata
  file. The resulting link file is written to ``STEP-NAME.KEYID-PREFIX.link``.

  Use this function together with in_toto_record_start as an alternative to
  in_toto_run, in order to provide evidence for supply chain steps that cannot
  be carried out by a single command.

  Arguments:
    step_name: A unique name to associate link metadata with a step.

    product_list: A list of artifact paths to be recorded as products.
        Directories are traversed recursively.

    gpg_keyid (optional): A keyid used to identify a local gpg key used to sign
        the resulting link metadata.

    gpg_use_default (optional): A boolean indicating if the default gpg key
        should be used to sign the resulting link metadata.

    gpg_home (optional): A path to the gpg home directory. If not set the
        default gpg home directory is used.

    exclude_patterns (optional): A list of filename patterns to exclude certain
        files from being recorded as artifacts.

    base_path (optional): A path relative to which artifacts are recorded.
        Default is the current working directory.

    normalize_line_endings (optional): A boolean indicating if line endings of
        artifacts should be normalized before hashing for cross-platform
        support.

    lstrip_paths (optional): A list of path prefixes used to left-strip
        artifact paths before storing them in the resulting link metadata.

    metadata_directory (optional): A directory path to write the resulting link
        metadata file to. Default destination is the current working directory.

    command (optional): A list consisting of a command and arguments executed
        between in_toto_record_start() and in_toto_record_stop() to capture
        the command ran in the resulting link metadata.

    byproducts (optional): A dictionary that lists byproducts of the link
        command execution. It should have at least the following entries
        "stdout" (str), "stderr" (str) and "return-value" (int).

    environment (optional): A dictionary to capture information about
        the environment to be added in the resulting link metadata eg.::

            {
              "variables": "<list of env var KEY=value pairs>",
              "filesystem": "<filesystem info>",
              "workdir": "<CWD when executing link command>"
            }

    signer (optional): A securesystemslib Signer instance used to
        sign the resulting link metadata.

  Raises:
    securesystemslib.exceptions.FormatError: Passed arguments are malformed.

    ValueError: None of gpg_keyid or gpg_use_default=True is
        passed.

    LinkNotFoundError: No preliminary link metadata file found.

    securesystemslib.exceptions.StorageError: Cannot hash artifacts.

    PrefixError: Left-stripping artifact paths results in non-unique dict keys.

    subprocess.TimeoutExpired: Link command times out.

    IOError, FileNotFoundError, NotADirectoryError, PermissionError:
        Cannot write link metadata.

    securesystemslib.exceptions.CryptoError, \
            securesystemslib.exceptions.UnsupportedAlgorithmError:
        Signing errors.

    ValueError, OSError, securesystemslib._gpg.exceptions.CommandError, \
            securesystemslib._gpg.exceptions.KeyNotFoundError:
        gpg signing errors.

  Side Effects:
    Reads preliminary link metadata file from disk.
    Reads artifact files from disk.
    Calls system gpg in a subprocess, if a gpg key argument is passed.
    Writes resulting link metadata file to disk.
    Removes preliminary link metadata file from disk.

  """
    # pylint: disable=too-many-branches, too-many-locals, too-many-statements
    LOG.info("Stop recording '%s'...", step_name)

    # Check that we have something to sign and if the formats are right
    _require_signing_arg(signer, gpg_keyid, gpg_use_default)

    if signer:
        _check_signer(signer)

    if gpg_keyid:
        _check_hex(gpg_keyid)

    if exclude_patterns:
        _check_str_list(exclude_patterns)

    if base_path:
        _check_str(base_path)

    if metadata_directory:
        _check_str(metadata_directory)

    # Load preliminary link file
    # If we have a signing key we can use the keyid to construct the name
    if signer:
        unfinished_fn = UNFINISHED_FILENAME_FORMAT.format(
            step_name=step_name, keyid=signer.public_key.keyid
        )

    # FIXME: Currently there is no way to know the default GPG key's keyid and
    # so we glob for preliminary link files
    else:
        unfinished_fn_glob = UNFINISHED_FILENAME_FORMAT_GLOB.format(
            step_name=step_name, pattern="*"
        )
        unfinished_fn_list = glob.glob(unfinished_fn_glob)

        if not unfinished_fn_list:
            raise in_toto.exceptions.LinkNotFoundError(
                "Could not find a preliminary"
                " link for step '{}' in the current working directory.".format(
                    step_name
                )
            )

        if len(unfinished_fn_list) > 1:
            raise in_toto.exceptions.LinkNotFoundError(
                "Found more than one"
                " preliminary links for step '{}' in the current working directory:"
                " {}. We need exactly one to stop recording.".format(
                    step_name, ", ".join(unfinished_fn_list)
                )
            )

        unfinished_fn = unfinished_fn_list[0]

    LOG.info("Loading preliminary link metadata '%s'...", unfinished_fn)
    link_metadata = Metadata.load(unfinished_fn)

    # The file must have been signed by the same key
    # If we have a signer we use it for verification as well
    if signer:
        LOG.info("Verifying preliminary link signature using passed signer...")
        keyid = signer.public_key.keyid
        verification_key = signer.public_key.to_dict()
        verification_key["keyid"] = keyid

    elif gpg_keyid:
        LOG.info("Verifying preliminary link signature using passed gpg key...")
        gpg_pubkey = securesystemslib._gpg.functions.export_pubkey(  # pylint: disable=protected-access
            gpg_keyid, gpg_home
        )
        keyid = gpg_pubkey["keyid"]
        verification_key = gpg_pubkey

    else:  # must be gpg_use_default
        # FIXME: Currently there is no way to know the default GPG key's keyid
        # before signing. As a workaround we extract the keyid of the preliminary
        # Link file's signature and try to export a pubkey from the gpg
        # home directory. We do this even if a gpg_keyid was specified, because gpg
        # accepts many different ids (mail, name, parts of an id, ...) but we
        # need a specific format.
        LOG.info(
            "Verifying preliminary link signature using default gpg key..."
        )

        # The `signatures` field is not part of the common Envelope/Metablock
        # interface, so we need to case handle. Note that we shouldn't be
        # accessing `signatures` here in the first place (see FIXME above).
        if isinstance(link_metadata, Envelope):
            keyid = link_metadata.signatures.values()[0].keyid

        else:
            keyid = link_metadata.signatures[0]["keyid"]

        gpg_pubkey = securesystemslib._gpg.functions.export_pubkey(  # pylint: disable=protected-access
            keyid, gpg_home
        )
        verification_key = gpg_pubkey

    link_metadata.verify_signature(verification_key)

    LOG.info("Extracting Link from metadata...")
    link = link_metadata.get_payload()

    # Record products if a product path list was passed
    if product_list:
        LOG.info("Recording products '%s'...", ", ".join(product_list))

    link.products = record_artifacts_as_dict(
        product_list,
        exclude_patterns=exclude_patterns,
        base_path=base_path,
        follow_symlink_dirs=True,
        normalize_line_endings=normalize_line_endings,
        lstrip_paths=lstrip_paths,
    )

    if command:
        link.command = command

    if byproducts:
        link.byproducts = byproducts

    if environment:
        link.environment = environment

    if isinstance(link_metadata, Metablock):
        LOG.info("Generating link metadata using Metablock...")
        link_metadata = Metablock(signed=link)
    else:
        LOG.info("Generating link metadata using DSSE...")
        link_metadata = Envelope.from_signable(link)

    if signer:
        LOG.info(
            "Updating signature with signer '{:.8}...'...".format(
                signer.public_key.keyid
            )
        )

    else:  # gpg_keyid or gpg_use_default
        # In both cases we use the keyid we got from verifying the preliminary
        # link signature above.
        LOG.info("Updating signature with gpg key '{:.8}...'...".format(keyid))
        signer = GPGSigner(keyid=keyid, homedir=gpg_home)

    link_metadata.create_signature(signer)
    fn = FILENAME_FORMAT.format(step_name=step_name, keyid=keyid)

    if metadata_directory is not None:
        fn = os.path.join(metadata_directory, fn)

    LOG.info("Storing link metadata to '%s'...", fn)
    link_metadata.dump(fn)

    LOG.info("Removing unfinished link metadata '%s'...", unfinished_fn)
    os.remove(unfinished_fn)


def in_toto_match_products(
    link, paths=None, exclude_patterns=None, lstrip_paths=None
):
    """Check if local artifacts match products in passed link.

    NOTE: Does not check integrity or authenticity of passed link!

    Arguments:
      link: The Link object to match.

    See ``in_toto_run`` for details about arguments, and exceptions that may
    occur while recording artifact hashes.

    Returns:
      A 3-tuple with artifact names that are
      - only in products,
      - not in products,
      - have different hashes.
    """
    if paths is None:
        paths = ["."]

    artifacts = record_artifacts_as_dict(
        paths, exclude_patterns=exclude_patterns, lstrip_paths=lstrip_paths
    )

    artifact_names = artifacts.keys()
    product_names = link.products.keys()

    only_products = product_names - artifact_names
    not_in_products = artifact_names - product_names
    differ = {
        name
        for name in product_names & artifact_names
        if link.products[name] != artifacts[name]
    }

    return only_products, not_in_products, differ



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Module Name>
  in_toto/rulelib.py

<Author>
  Santiago Torres-Arias <santiago@nyu.edu>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  Nov 18, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  This module provides functions parse artifact rules and validate their
  syntax.

"""
import securesystemslib.exceptions
import securesystemslib.formats

from in_toto.formats import _check_str, _check_str_list

GENERIC_RULES = {
    "create",
    "modify",
    "delete",
    "allow",
    "disallow",
    "require",
}
COMPLEX_RULES = {
    "match",
}
ALL_RULES = GENERIC_RULES | COMPLEX_RULES


def unpack_rule(rule):
    """
    Parses the rule and extracts and returns the necessary data to apply the
    rule. Can also be used to verify if a rule complies with any of the formats

    <Arguments>
      rule:
          The list of rule elements, in one of the following formats:
          MATCH <pattern> [IN <source-path-prefix>] WITH (MATERIALS|PRODUCTS)
              [IN <destination-path-prefix>] FROM <step>,
          CREATE <pattern>,
          DELETE <pattern>,
          MODIFY <pattern>,
          ALLOW <pattern>,
          DISALLOW <pattern>,
          REQUIRE <file>

    Note that REQUIRE is somewhat of a weird animal that does not use patterns
    but rather single filenames (for now).

    <Exceptions>
      raises FormatError, if the rule does not comply with any of the formats.

    <Side Effects>
      None.

    <Returns>
      A dictionary of the artifact rule data,
      if it is a generic rule the dictionary is:
      {
        "rule_type": rule[0] ("CREATE"|"MODIFY"|"DELETE"|"ALLOW"|"DISALLOW")
        "pattern" : rule[1], a path pattern
      }

      if it is a match rule, the dictionary is:
      {
        "rule_type": rule[0],  ("MATCH"),
        "pattern": rule[1], a path pattern
        "source_prefix": path or empty string
        "dest_prefix": path or empty string
        "dest_type" : destination artifact type, ("MATERIAL"|"PRODUCT")
        "dest_name": destination step/inspection name
      }

    """
    _check_str_list(rule)

    # Create all lower rule copy to case insensitively parse out tokens whose
    # position we don't know yet
    # We keep the original rule to retain the non-token elements' case
    rule_lower = []
    for rule_elem in rule:
        rule_lower.append(rule_elem.lower())

    rule_len = len(rule)

    if rule_len < 2 or rule_lower[0] not in ALL_RULES:
        raise securesystemslib.exceptions.FormatError(
            "Wrong rule format,"
            " rules must start with one of '{0}' and specify a 'pattern' as"
            " second element.\n"
            "Got: \n\t'{1}'".format(", ".join(ALL_RULES), rule)
        )

    rule_type = rule_lower[0]
    pattern = rule[1]

    # Type is one of "CREATE", "MODIFY", "DELETE", "ALLOW", "DISALLOW"
    if rule_type in GENERIC_RULES:
        if rule_len != 2:
            raise securesystemslib.exceptions.FormatError(
                "Wrong rule format,"
                " generic rules must have one of the formats:\n\t"
                "CREATE <pattern>\n\t"
                "MODIFY <pattern>\n\t"
                "DELETE <pattern>\n\t"
                "ALLOW <pattern>\n\t"
                "DISALLOW <pattern>\n"
                "REQUIRE <file>\n"
                "Got:\n\t{}".format(rule)
            )
        return {
            "rule_type": rule_type,
            "pattern": pattern,
        }

    # Type is "MATCH"
    if rule_type in COMPLEX_RULES:
        # ... IN <source-path-prefix> WITH (MATERIALS|PRODUCTS)
        # IN <destination-path-prefix> FROM <step>
        if (
            rule_len == 10
            and rule_lower[2] == "in"
            and rule_lower[4] == "with"
            and rule_lower[6] == "in"
            and rule_lower[8] == "from"
        ):
            source_prefix = rule[3]
            dest_type = rule_lower[5]
            dest_prefix = rule[7]
            dest_name = rule[9]

        # ... IN <source-path-prefix> WITH (MATERIALS|PRODUCTS) FROM <step>
        elif (
            rule_len == 8
            and rule_lower[2] == "in"
            and rule_lower[4] == "with"
            and rule_lower[6] == "from"
        ):
            source_prefix = rule[3]
            dest_type = rule_lower[5]
            dest_prefix = ""
            dest_name = rule[7]

        # ... WITH (MATERIALS|PRODUCTS) IN <destination-path-prefix> FROM <step>
        elif (
            rule_len == 8
            and rule_lower[2] == "with"
            and rule_lower[4] == "in"
            and rule_lower[6] == "from"
        ):
            source_prefix = ""
            dest_type = rule_lower[3]
            dest_prefix = rule[5]
            dest_name = rule[7]

        # ... WITH (MATERIALS|PRODUCTS) FROM <step>
        elif (
            rule_len == 6
            and rule_lower[2] == "with"
            and rule_lower[4] == "from"
        ):
            source_prefix = ""
            dest_type = rule_lower[3]
            dest_prefix = ""
            dest_name = rule[5]

        else:
            raise securesystemslib.exceptions.FormatError(
                "Wrong rule format,"
                " match rules must have the format:\n\t"
                " MATCH <pattern> [IN <source-path-prefix>] WITH"
                " (MATERIALS|PRODUCTS) [IN <destination-path-prefix>] FROM <step>.\n"
                "Got: \n\t{}".format(rule)
            )

        if dest_type not in {"materials", "products"}:
            raise securesystemslib.exceptions.FormatError(
                "Wrong rule format,"
                " match rules must have either MATERIALS or PRODUCTS (case"
                " insensitive) as destination.\n"
                "Got: \n\t{}".format(rule)
            )

        return {
            "rule_type": rule_type,
            "pattern": pattern,
            "source_prefix": source_prefix,
            "dest_prefix": dest_prefix,
            "dest_type": dest_type,
            "dest_name": dest_name,
        }
    return None


def pack_rule(
    rule_type,
    pattern,
    source_prefix=None,
    dest_type=None,
    dest_prefix=None,
    dest_name=None,
):
    """
    <Purpose>
      Create an artifact rule in the passed arguments and return it as list
      as it is stored in a step's or inspection's expected_material or
      expected_product field.

    <Arguments>
      rule_type:
              One of "MATCH", "CREATE", "DELETE", MODIFY, ALLOW, DISALLOW,
              REQUIRE (case insensitive).

      pattern:
              A glob pattern to match artifact paths.

      source_prefix: (only for "MATCH" rules)
              A prefix for 'pattern' to match artifacts reported by the link
              corresponding to the step that contains the rule.

      dest_type: (only for "MATCH" rules)
              One of "MATERIALS" or "PRODUCTS" (case insensitive) to specify
              if materials or products of the link corresponding to the step
              identified by 'dest_name' should be matched.

      dest_prefix: (only for "MATCH" rules)
              A prefix for 'pattern' to match artifacts reported by the link
              corresponding to the step identified by 'dest_name'.

      dest_name: (only for "MATCH" rules)
              The name of the step whose corresponding link is used to match
              artifacts.


    <Exceptions>
      securesystemslib.exceptions.FormatError
              if any of the arguments is malformed.


    <Returns>
      One of the following rule formats in as a list
        MATCH <pattern> [IN <source-path-prefix>] WITH (MATERIALS|PRODUCTS)
            [IN <destination-path-prefix>] FROM <step>,
        CREATE <pattern>,
        DELETE <pattern>,
        MODIFY <pattern>,
        ALLOW <pattern>,
        DISALLOW <pattern>,
        REQUIRE <file>

      Note that REQUIRE is somewhat of a weird animal that does not use patterns
      but rather single filenames (for now).
    """
    _check_str(rule_type)
    _check_str(pattern)

    if rule_type.lower() not in ALL_RULES:
        raise securesystemslib.exceptions.FormatError(
            "'{0}' is not a valid "
            "'type'.  Rule type must be one of:  {1} (case insensitive).".format(
                rule_type, ", ".join(ALL_RULES)
            )
        )

    if rule_type.upper() == "MATCH":
        if not isinstance(dest_type, str) or not (
            dest_type.lower() == "materials" or dest_type.lower() == "products"
        ):
            raise securesystemslib.exceptions.FormatError(
                "'{}' is not a valid"
                " 'dest_type'. Rules of type 'MATCH' require a destination type of"
                " either 'MATERIALS' or 'PRODUCTS' (case insensitive).".format(
                    dest_type
                )
            )

        if not (isinstance(dest_name, str) and dest_name):
            raise securesystemslib.exceptions.FormatError(
                "'{}' is not a valid"
                " 'dest_name'. Rules of type 'MATCH' require a step name as a"
                " destination name.".format(dest_name)
            )

        # Construct rule
        rule = ["MATCH", pattern]

        if source_prefix:
            _check_str(source_prefix)

            rule += ["IN", source_prefix]

        rule += ["WITH", dest_type.upper()]

        if dest_prefix:
            _check_str(dest_prefix)
            rule += ["IN", dest_prefix]

        rule += ["FROM", dest_name]

    else:
        rule = [rule_type.upper(), pattern]

    return rule


def pack_rule_data(rule_data):
    """Shortcut for 'pack_rule' to pack a rule based on a rule_data dictionary
    as returned by 'unpack_rule'."""
    return pack_rule(**rule_data)


def pack_create_rule(pattern):
    """Shortcut for 'pack_rule' to pack a CREATE rule."""
    return pack_rule("CREATE", pattern)


def pack_delete_rule(pattern):
    """Shortcut for 'pack_rule' to pack a DELETE rule."""
    return pack_rule("DELETE", pattern)


def pack_modify_rule(pattern):
    """Shortcut for 'pack_rule' to pack a MODIFY rule."""
    return pack_rule("MODIFY", pattern)


def pack_allow_rule(pattern):
    """Shortcut for 'pack_rule' to pack an ALLOW rule."""
    return pack_rule("ALLOW", pattern)


def pack_disallow_rule(pattern):
    """Shortcut for 'pack_rule' to pack a DISALLOW rule."""
    return pack_rule("DISALLOW", pattern)


def pack_require_rule(filename):
    """
    Shortcut for 'pack_rule' to pack a REQUIRE rule:
    note that REQUIRE is somewhat of a weird animal that does not use patterns
    but rather single filenames (for now).
    """
    return pack_rule("REQUIRE", filename)



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
Configure base logger for in_toto (see in_toto.log for details).

"""
import in_toto.log

# in-toto version
__version__ = "3.0.0"



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  common.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>
  Santiago Torres <santiago@nyu.edu>

<Started>
  Sep 23, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides base classes for various classes in the model.

"""

import inspect
import json

import attr
import securesystemslib.formats


class ValidationMixin:
    """The validation mixin provides a self-inspecting method, validate, to
    allow in-toto's objects to check that they are proper."""

    def validate(self):
        """Validates attributes of the instance.

        Raises:
          securesystemslib.formats.FormatError: An attribute value is invalid.

        """
        for method in inspect.getmembers(self, predicate=inspect.ismethod):
            if method[0].startswith("_validate_"):
                method[1]()


@attr.s(repr=False, init=False)
class Signable(ValidationMixin):
    """Objects with base class Signable are to be included in a Metablock class
    to be signed (hence the name). They provide a `signable_bytes` property
    used to create deterministic signatures."""

    def __repr__(self):
        """Returns an indented JSON string of the metadata object."""
        return json.dumps(
            attr.asdict(self), indent=1, separators=(",", ": "), sort_keys=True
        )

    @property
    def signable_bytes(self):
        """The UTF-8 encoded canonical JSON byte representation of the dictionary
        representation of the instance."""
        return securesystemslib.formats.encode_canonical(
            attr.asdict(self)
        ).encode("UTF-8")



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  _signer.py

<Author>
  Pradyumna Krishna <git@onpy.in>

<Started>
  Jan 26, 2023

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides in-toto flavored Signer implementations

"""

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import securesystemslib._gpg.exceptions as gpg_exceptions
import securesystemslib._gpg.functions as gpg
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key,
)
from securesystemslib import exceptions
from securesystemslib.signer import (
    CryptoSigner,
    Key,
    SecretsHandler,
    Signature,
    Signer,
    SSlibKey,
)


def load_crypto_signer_from_pkcs8_file(
    path: str, password: Optional[bytes] = None
) -> CryptoSigner:
    """Internal helper to load CryptoSigner from PKCS8/PEM file."""
    with open(path, "rb") as f:
        data = f.read()

    private_key = load_pem_private_key(data, password)
    signer = CryptoSigner(private_key)

    return signer


def load_public_key_from_file(path: str) -> Dict[str, Any]:
    """Internal helper to load key from SubjectPublicKeyInfo/PEM file."""
    with open(path, "rb") as f:
        data = f.read()

    crypto_public_key = load_pem_public_key(data)
    key = SSlibKey.from_crypto(crypto_public_key)

    # NOTE: Would be nice to support `Key` instances natively in in-toto.
    # Until then we keep using the in-toto -tailored dict representation of
    # `Key`, which is expected in metadata model and verifylib API.
    key_dict = key.to_dict()
    key_dict["keyid"] = key.keyid

    return key_dict


class GPGSignature(Signature):
    """A container class containing information about a gpg signature.
    Besides the signature, it also contains other meta information
    needed to uniquely identify the key used to generate the signature.

    Attributes:
      keyid: HEX string used as a unique identifier of the key.
      signature: HEX string representing the signature.
      other_headers: HEX representation of additional GPG headers.
    """

    def __init__(
        self,
        keyid: str,
        signature: str,
        other_headers: str,
    ):
        super().__init__(keyid, signature)
        self.other_headers = other_headers

    @classmethod
    def from_dict(cls, signature_dict: Dict) -> "GPGSignature":
        """Creates a ``GPGSignature`` object from its JSON/dict
        representation.

        Arguments:
          signature_dict: Dict containing valid "keyid", "signature" and
            "other_fields" fields.
        Raises:
          KeyError: If any of the "keyid", "sig" or "other_headers" fields
            are missing from the signature_dict.
        Returns:
          ``GPGSignature`` instance.
        """

        return cls(
            signature_dict["keyid"],
            signature_dict["signature"],
            signature_dict["other_headers"],
        )

    def to_dict(self) -> Dict:
        """Returns the JSON-serializable dictionary representation of self."""
        return {
            "keyid": self.keyid,
            "signature": self.signature,
            "other_headers": self.other_headers,
        }


class GPGSigner(Signer):
    """A in-toto gpg implementation of the ``Signer`` interface.
    Provides a sign method to generate a cryptographic signature with gpg, using
    an RSA, DSA or EdDSA private key identified by the keyid on the instance.

    Arguments:
      keyid: The keyid of the gpg signing keyid. If not passed the default
        key in the keyring is used.
      homedir: Path to the gpg keyring. If not passed the default keyring
        is used.
    """

    def __init__(
        self, keyid: Optional[str] = None, homedir: Optional[str] = None
    ):
        self.keyid = keyid
        self.homedir = homedir

    @property
    def public_key(self) -> Key:
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def from_priv_key_uri(
        cls,
        priv_key_uri: str,
        public_key: Key,
        secrets_handler: Optional[SecretsHandler] = None,
    ) -> "GPGSigner":
        raise NotImplementedError(
            "Incompatible with private key URIs"
        )  # pragma: no cover

    def sign(self, payload: bytes) -> GPGSignature:
        """Signs a given payload by the key assigned to the ``GPGSigner``
        instance. Calls the gpg command line utility to sign the passed content
        with the key identified by the passed keyid from the gpg keyring at the
        passed homedir.

        Arguments:
          payload: The bytes to be signed.
        Raises:
          securesystemslib.exceptions.FormatError:
            If the keyid is invalid.
          ValueError: the gpg command failed to create a valid signature.
          OSError: the gpg command is not present or non-executable.
          securesystemslib.exceptions.UnsupportedLibraryError: the gpg command is
            not available, or the cryptography library is not installed.
          securesystemslib._gpg.exceptions.CommandError: the gpg command returned a
            non-zero exit code.
          securesystemslib._gpg.exceptions.KeyNotFoundError: the used gpg version is
            not fully supported and no public key can be found for short keyid.
          Returns:
            Returns a ``GPGSignature`` class instance.
        """

        sig_dict = gpg.create_signature(payload, self.keyid, self.homedir)
        return GPGSignature(**sig_dict)


@dataclass
class GPGKey(Key):
    """A container class representing public key portion of a GPG key.
    Provides a verify method to verify a cryptographic signature with a
    gpg-style rsa, dsa or ecdsa public key on the instance.

    Attributes:
      type: Key type, e.g. "rsa", "dsa" or "ecdsa".
      method: GPG Key Scheme, For example:
        "pgp+rsa-pkcsv1.5", "pgp+dsa-fips-180-2", and "pgp+eddsa-ed25519".
      hashes: list of GPG Hash Algorithms, e.g. "pgp+SHA2".
      keyval: Opaque key content.
      keyid: Key identifier that is unique within the metadata it is used in.
        Keyid is not verified to be the hash of a specific representation
        of the key.
      creation_time: Unix timestamp when GPG key was created.
      validity_period: Validity of the GPG Keys in days.
      subkeys: A dictionary containing keyid and GPG subkey.
    """

    # pylint: disable=too-many-instance-attributes

    type: str
    method: str
    hashes: List[str]
    keyval: Dict[str, str]
    keyid: str
    creation_time: Optional[int] = None
    validity_period: Optional[int] = None
    subkeys: Optional[Dict[str, "GPGKey"]] = None

    @classmethod
    def from_dict(cls, keyid: str, key_dict: Dict[str, Any]):
        """Creates ``GPGKey`` object from its json/dict representation.
        Raises:
          KeyError, TypeError: Invalid arguments.
        """
        subkeys_dict = key_dict.get("subkeys")

        gpg_subkeys = None
        if subkeys_dict:
            gpg_subkeys = {
                keyid_: GPGKey.from_dict(keyid_, subkey_dict)
                for (keyid_, subkey_dict) in subkeys_dict.items()
            }

        return cls(
            key_dict["type"],
            key_dict["method"],
            key_dict["hashes"],
            key_dict["keyval"],
            keyid,
            key_dict.get("creation_time"),
            key_dict.get("validity_period"),
            gpg_subkeys,
        )

    @classmethod
    def from_legacy_dict(cls, key_dict: Dict[str, Any]):
        """Create GPGKey from legacy dictionary representation."""

        keyid = key_dict["keyid"]
        return cls.from_dict(keyid, key_dict)

    def to_dict(self):
        """Returns the dictionary representation of self."""

        key_dict = {
            "method": self.method,
            "type": self.type,
            "hashes": self.hashes,
            "keyid": self.keyid,
            "keyval": self.keyval,
        }

        if self.creation_time:
            key_dict["creation_time"] = self.creation_time
        if self.validity_period:
            key_dict["validity_period"] = self.validity_period
        if self.subkeys:
            subkeys_dict = {
                keyid: subkey.to_dict()
                for (keyid, subkey) in self.subkeys.items()
            }
            key_dict["subkeys"] = subkeys_dict

        return key_dict

    @classmethod
    def from_keyring(cls, keyid, homedir=None):
        """Creates ``GPGKey`` object from GnuPG Keyring."""

        pubkey_dict = gpg.export_pubkey(keyid, homedir)
        return cls.from_dict(keyid, pubkey_dict)

    def verify_signature(self, signature: GPGSignature, data: bytes) -> None:
        """Verifies a given payload by the key assigned to the GPGKey
        instance.

        Arguments:
          signature: A ``GPGSignature`` class instance.
          data: The bytes to be verified.
        """

        try:
            if not gpg.verify_signature(
                signature.to_dict(), self.to_dict(), data
            ):
                raise exceptions.UnverifiedSignatureError(
                    f"Failed to verify signature by {self.keyid}"
                )
        except (
            exceptions.FormatError,
            exceptions.UnsupportedLibraryError,
            gpg_exceptions.KeyExpirationError,
        ) as e:
            raise exceptions.VerificationError(
                f"Unknown failure to verify signature by {self.keyid}"
            ) from e



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  formats.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>
  Santiago Torres-Arias <santiago@nyu.edu>

<Started>
  November 28, 2017.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Helpers to validate API inputs and metadata model objects.

"""
from copy import deepcopy
from re import fullmatch

from securesystemslib.exceptions import FormatError
from securesystemslib.signer import Key, Signature

from in_toto.models._signer import GPGKey, GPGSignature


def _err(arg, expected):
    return FormatError(f"expected {expected}, got '{arg} ({type(arg)})'")


def _check_int(arg):
    if not isinstance(arg, int):
        raise _err(arg, "int")


def _check_str(arg):
    if not isinstance(arg, str):
        raise _err(arg, "str")


def _check_hex(arg):
    _check_str(arg)
    if fullmatch(r"^[0-9a-fA-F]+$", arg) is None:
        raise _err(arg, "hex string")


def _check_list(arg):
    if not isinstance(arg, list):
        raise _err(arg, "list")


def _check_dict(arg):
    if not isinstance(arg, dict):
        raise _err(arg, "dict")


def _check_str_list(arg):
    _check_list(arg)
    for e in arg:
        _check_str(e)


def _check_hex_list(arg):
    _check_list(arg)
    for e in arg:
        _check_hex(e)


def _check_iso8601(arg):
    """Check iso8601 date format."""
    _check_str(arg)
    if fullmatch(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z$", arg) is None:
        raise _err(arg, "'YYYY-MM-DDTHH:MM:SSZ'")


def _check_hash_dict(arg):
    """Check artifact hash dict."""
    _check_dict(arg)
    for k, v in arg.items():
        _check_str(k)
        _check_hex(v)


def _check_parameter_dict(arg):
    """Check verifylib parameter dict."""
    _check_dict(arg)
    for k, v in arg.items():
        _check_str(k)
        if fullmatch(r"^[a-zA-Z0-9_-]+$", k) is None:
            raise _err(arg, "'a-zA-Z0-9_-'")
        _check_str(v)


def _check_signature(arg):
    """Check signature dict."""
    _check_dict(arg)
    # NOTE: `GPGSignature` and `Signature` serialization formats are incompatible
    try:
        GPGSignature.from_dict(arg)
    except KeyError:
        try:
            Signature.from_dict(deepcopy(arg))
        except KeyError as e:
            raise _err(arg, "signature dict") from e


def _check_public_key(arg):
    """Check public key dict."""
    _check_dict(arg)
    # NOTE: `GPGKey` and `Key` serialization formats are incompatible
    try:
        GPGKey.from_dict(arg["keyid"], arg)
    except (KeyError, TypeError):
        try:
            Key.from_dict(arg["keyid"], deepcopy(arg))
        except KeyError as e:
            raise _err(arg, "public key dict") from e


def _check_public_keys(arg):
    """Check dict of public key dicts."""
    _check_dict(arg)
    for k, v in arg.items():
        _check_hex(k)
        _check_public_key(v)



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

# TODO: Add module docstring and remove pylint exemption in in-toto/in-toto#126
# pylint: disable=missing-docstring
from securesystemslib.exceptions import Error


class SignatureVerificationError(Error):
    """Indicates a signature verification Error."""


class LayoutExpiredError(Error):
    """Indicates that the layout expired."""


class RuleVerificationError(Error):
    """Indicates that artifact rule verification failed."""


class ThresholdVerificationError(Error):
    """Indicates that signature threshold verification failed."""


class BadReturnValueError(Error):
    """Indicates that a ran command exited with non-int or non-zero return
    value."""


class LinkNotFoundError(Error):
    """Indicates that a link file was not found."""


class UnsupportedKeyTypeError(Error):
    """Indicates that the specified key type is not yet supported."""


class PrefixError(Error):
    """Indicates that there is an error because of the prefixes passed."""


class InvalidMetadata(Error):
    """Indicates that the metadata is not valid."""



"""Artifact resolver API.

Extensible interface to hash artifacts based on URIs.

Example usage::

    from in_toto.resolver import Resolver, RESOLVER_FOR_URI_SCHEME

    # Register resolver instances for schemes
    RESOLVER_FOR_URI_SCHEME["myscheme"] = MyResolver()

    # Resolve arbitrary URIs with registered resolvers
    resolver = Resolver.for_uri(uri)
    artifact_hashes = resolver.hash_artifacts(uris)

"""

from in_toto.resolver._resolver import (
    RESOLVER_FOR_URI_SCHEME,
    DirectoryResolver,
    FileResolver,
    OSTreeResolver,
    Resolver,
)



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  metadata.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>
  Santiago Torres <santiago@nyu.edu>

<Started>
  Oct 23, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides a container class `Metablock` for signed metadata and
  functions for signing, signature verification, de-serialization and
  serialization from and to JSON.

"""

import json
from copy import deepcopy
from typing import Union

import attr
import securesystemslib._gpg.functions
import securesystemslib.exceptions
import securesystemslib.formats
from securesystemslib.dsse import Envelope as SSlibEnvelope
from securesystemslib.exceptions import (
    UnverifiedSignatureError,
    VerificationError,
)
from securesystemslib.signer import Key, Signature, Signer

from in_toto.exceptions import InvalidMetadata, SignatureVerificationError
from in_toto.formats import _check_public_key, _check_signature
from in_toto.models._signer import GPGSigner
from in_toto.models.common import Signable, ValidationMixin
from in_toto.models.layout import Layout
from in_toto.models.link import Link

ENVELOPE_PAYLOAD_TYPE = "application/vnd.in-toto+json"


class Metadata:
    """A Metadata abstraction between DSSE Envelope and Metablock."""

    @classmethod
    def from_dict(cls, data):
        """Loads DSSE or Traditional Metadata from its JSON/dict representation."""

        if "payload" in data:
            if data.get("payloadType") == ENVELOPE_PAYLOAD_TYPE:
                return Envelope.from_dict(data)

        elif "signed" in data:
            return Metablock.from_dict(data)

        raise InvalidMetadata

    def to_dict(self):
        """Returns the JSON-serializable dictionary representation of self."""
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def load(cls, path):
        """Loads the JSON string representation of metadata from disk.

        Arguments:
          path: The path to read the file from.

        Raises:
          IOError: The file cannot be read.
          InvalidMetadata: Metadata format is invalid.
          securesystemslib.exceptions.FormatError: Metadata format is invalid.

        Returns:
          A Metadata containing a Link or Layout object.

        """
        with open(path, "r", encoding="utf8") as fp:
            data = json.load(fp)

        return cls.from_dict(data)

    def dump(self, path):
        """Writes the JSON string representation of the instance to disk.

        Arguments:
          path: The path to write the file to.

        Raises:
          IOError: File cannot be written.

        """
        json_bytes = json.dumps(
            self.to_dict(),
            sort_keys=True,
        ).encode("utf-8")

        with open(path, "wb") as fp:
            fp.write(json_bytes)

    def create_signature(self, signer: Signer) -> Signature:
        """Creates and adds signature over signable representation of self.

        The passed signer is used for signing. Applications can implement their
        own signer or use one from ``securesystemslib``.

        Arguments:
            signer: A ``Signer`` implementation.

        Returns:
            The ``Signature`` object returned from ``Signer``.

        """
        raise NotImplementedError  # pragma: no cover

    def verify_signature(self, verification_key):
        """Verifies a signature over signable in signatures with verification_key.

        Arguments:
          verification_key: A verification key.

        Raises:
          securesystemslib.exceptions.FormatError: The passed key is malformed.

          SignatureVerificationError: No signature keyid matches
              the verification key keyid, or the matching signature is malformed,
              or the matching signature is invalid.
        """
        raise NotImplementedError  # pragma: no cover

    def get_payload(self):
        """Returns ``Link`` or ``Layout``."""
        raise NotImplementedError  # pragma: no cover


class Envelope(SSlibEnvelope, Metadata):
    """DSSE Envelope for in-toto payloads."""

    @classmethod
    def from_signable(cls, signable: Signable) -> "Envelope":
        """Creates DSSE envelope with signable bytes as payload."""

        json_bytes = json.dumps(
            attr.asdict(signable),
            sort_keys=True,
        ).encode("utf-8")

        return cls(
            payload=json_bytes,
            payload_type=ENVELOPE_PAYLOAD_TYPE,
            signatures={},
        )

    def create_signature(self, signer: Signer) -> Signature:
        if isinstance(signer, GPGSigner):
            raise NotImplementedError("GPG Signing is not implemented")

        return super().sign(signer)

    def verify_signature(self, verification_key):
        # Deepcopy to preserve `verification_key`, which might still be needed
        # in calling context and would otherwise be destroyed in `from_dict`.
        # NOTE: It would be nice to support `Key` natively in in-toto model.
        key = Key.from_dict(
            verification_key["keyid"], deepcopy(verification_key)
        )

        try:
            super().verify(keys=[key], threshold=1)
        except VerificationError as exc:
            raise SignatureVerificationError from exc

    def get_payload(self) -> Union[Link, Layout]:
        """Parse DSSE payload into Link or Layout object.

        Raises:
            InvalidMetadata: If type in payload is not ``link`` or ``layout``.

        Returns:
            Link or Layout.
        """

        data = json.loads(self.payload.decode("utf-8"))
        _type = data.get("_type")
        if _type == "link":
            return Link.read(data)
        if _type == "layout":
            return Layout.read(data)

        raise InvalidMetadata


@attr.s(repr=False, init=False)
class Metablock(Metadata, ValidationMixin):
    """A container for signed in-toto metadata.

    Provides methods for metadata JSON (de-)serialization, reading from and
    writing to disk, creating and verifying signatures, and self-validation.

    Attributes:
      signed: A subclass of Signable which has the actual metadata payload,
          usually a Link or Layout object.
      signatures: A list of signatures over the canonical JSON representation
          of the value of the signed attribute.
      compact_json: A boolean indicating if the dump method should write a
          compact JSON string representation of the metadata.

    """

    signatures = attr.ib()
    signed = attr.ib()

    def __init__(self, **kwargs):
        self.signatures = kwargs.get("signatures", [])
        self.signed = kwargs.get("signed")
        self.compact_json = kwargs.get("compact_json", False)

        self.validate()

    def __repr__(self):
        """Returns the JSON string representation."""
        indent = None if self.compact_json else 1
        separators = (",", ":") if self.compact_json else (",", ": ")

        return json.dumps(
            {"signatures": self.signatures, "signed": attr.asdict(self.signed)},
            indent=indent,
            separators=separators,
            sort_keys=True,
        )

    def dump(self, path):
        """Writes the JSON string representation of the instance to disk.

        Arguments:
          path: The path to write the file to.

        Raises:
          IOError: File cannot be written.

        """
        with open(path, "wb") as fp:
            fp.write("{}".format(self).encode("utf-8"))

    @classmethod
    def from_dict(cls, data):
        """Creates a Metablock object from its JSON/dict representation."""

        signatures = data.get("signatures", [])
        signed_data = data.get("signed", {})
        signed_type = signed_data.get("_type")

        if signed_type == "link":
            signed = Link.read(signed_data)

        elif signed_type == "layout":
            signed = Layout.read(signed_data)

        else:
            raise securesystemslib.exceptions.FormatError(
                "Invalid Metadata format"
            )

        return cls(signatures=signatures, signed=signed)

    def to_dict(self):
        """Returns the JSON-serializable dictionary representation of self."""

        return {
            "signatures": self.signatures,
            "signed": attr.asdict(self.signed),
        }

    @property
    def type_(self):
        """A shortcut to the `type_` attribute of the object on the signable
        attribute (should be one of "link" or "layout")."""
        # NOTE: Trailing underscore is used by convention (pep8) to avoid conflict
        # with Python's type keyword.
        return self.signed.type_

    def create_signature(self, signer: Signer):
        signature = signer.sign(self.signed.signable_bytes)
        self.signatures.append(signature.to_dict())

        return signature

    def sign_gpg(self, gpg_keyid=None, gpg_home=None):
        """Creates signature over signable with gpg and adds it to signatures.

    Uses the UTF-8 encoded canonical JSON byte representation of the signable
    attribute to create signatures deterministically.

    Arguments:
      gpg_keyid (optional): A keyid used to identify a local gpg signing key.
          If omitted the default signing key is used.

      gpg_home (optional): A path to the gpg home directory. If not set the
          default gpg home directory is used.

    Raises:
      ValueError, OSError, securesystemslib._gpg.exceptions.CommandError, \
            securesystemslib._gpg.exceptions.KeyNotFoundError:
        gpg signing errors.

    Side Effects:
      Calls system gpg command in a subprocess.

    Returns:
      The signature.

    """
        signature = securesystemslib._gpg.functions.create_signature(  # pylint: disable=protected-access
            self.signed.signable_bytes, gpg_keyid, gpg_home
        )

        self.signatures.append(signature)

        return signature

    def verify_signature(self, verification_key):
        """Verifies a signature over signable in signatures with verification_key.

        Uses the UTF-8 encoded canonical JSON byte representation of the signable
        attribute to verify the signature deterministically.

        NOTE: Only the first signature in the signatures attribute, whose keyid
        matches the verification_key keyid, is verified. If the verification_key
        is a gpg key, subkey keyids are considered too.

        Arguments:
          verification_key: A verification key.

        Raises:
          securesystemslib.exceptions.FormatError: The passed key is malformed.

          SignatureVerificationError: No signature keyid matches the verification
              key keyid, or the matching signature is malformed, or the matching
              signature is invalid.

          securesystemslib._gpg.exceptions.KeyExpirationError: Passed verification
              key is an expired gpg key.

        """
        _check_public_key(verification_key)
        verification_keyid = verification_key["keyid"]

        # Find a signature that corresponds to the keyid of the passed
        # verification key or one of its subkeys
        signature = None
        for signature in self.signatures:
            if signature["keyid"] == verification_keyid:
                break

            if signature["keyid"] in list(
                verification_key.get("subkeys", {}).keys()
            ):
                break

        else:
            raise SignatureVerificationError(
                "No signature found for key '{}'".format(verification_keyid)
            )

        valid = False
        if "signature" in signature and "other_headers" in signature:
            valid = securesystemslib._gpg.functions.verify_signature(  # pylint: disable=protected-access
                signature, verification_key, self.signed.signable_bytes
            )

        else:
            # Parse key and (below) signature dicts as `Key` and `Signature`
            # instances to use modern securesystemslib verification code.
            # Deepcopy to preserve original dicts, which are otherwise destroyed
            # in `from_dict` methods.
            # NOTE: It would be nice to support `Key` and `Signature` natively
            # in in-toto model classes.
            key = Key.from_dict(verification_keyid, deepcopy(verification_key))

            try:
                sig = Signature.from_dict(deepcopy(signature))
                key.verify_signature(sig, self.signed.signable_bytes)
                valid = True

            except (KeyError, UnverifiedSignatureError):
                pass

        if not valid:
            raise SignatureVerificationError(
                "Invalid signature for keyid '{}'".format(verification_keyid)
            )

    def _validate_signed(self):
        """Private method to check if the 'signed' attribute contains a valid
        Layout or Link object."""

        if not isinstance(self.signed, (Layout, Link)):
            raise securesystemslib.exceptions.FormatError(
                "The Metblock's 'signed'"
                " property has has to be of type 'Link' or 'Layout'."
            )

        # If the signed object is a Link or Layout object validate it.
        self.signed.validate()

    def _validate_signatures(self):
        """Private method to check that the 'signatures' attribute is valid."""

        if not isinstance(self.signatures, list):
            raise securesystemslib.exceptions.FormatError(
                "The Metablock's"
                " 'signatures' property has to be of type 'list'."
            )

        for signature in self.signatures:
            _check_signature(signature)

    def get_payload(self):
        """Returns signed of the Metablock."""

        return self.signed



#!/usr/bin/env python

# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  link.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>
  Santiago Torres <santiago@nyu.edu>

<Started>
  Sep 23, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides a class for link metadata which is information gathered when a
  step of the supply chain is performed.
"""

import attr
import securesystemslib.formats

from in_toto.formats import _check_hash_dict
from in_toto.models.common import Signable

FILENAME_FORMAT = "{step_name}.{keyid:.8}.link"
FILENAME_FORMAT_SHORT = "{step_name}.link"
UNFINISHED_FILENAME_FORMAT = ".{step_name}.{keyid:.8}.link-unfinished"
UNFINISHED_FILENAME_FORMAT_GLOB = ".{step_name}.{pattern}.link-unfinished"


@attr.s(repr=False, init=False)
class Link(Signable):
    """Evidence for a performed step or inspection of the supply chain.

    A Link object is usually contained in a generic Metablock object for signing,
    serialization and I/O capabilities.

    Attributes:
      name: A unique name used to identify the related step or inspection in an
          in-toto layout.

      command: A list of command and command arguments that report how the
          corresponding step is performed.

      materials: A dictionary of the artifacts *used* by the step, i.e::

              {
                "<material path>": {
                  "<hash algorithm name>": "<hash digest of material>",
                  ...
                },
                ...
              }

      products: A dictionary of the artifacts *produced* by the step, i.e::

              {
                "<product path>": {
                  "<hash algorithm name>": "<hash digest of product>",
                  ...
                },
                ...
              }

      byproducts: An opaque dictionary that lists byproducts of the link command
          execution. It should have at least the following entries
          "stdout" (str), "stderr" (str) and "return-value" (int).

      environment: An opaque dictionary that lists information about the
          execution environment of the link command. eg.::

              {
                "variables": "<list of env var KEY=value pairs>",
                "filesystem": "<filesystem info>",
                "workdir": "<CWD when executing link command>"
              }

    """

    _type = attr.ib()
    name = attr.ib()
    materials = attr.ib()
    products = attr.ib()
    byproducts = attr.ib()
    command = attr.ib()
    environment = attr.ib()

    def __init__(self, **kwargs):
        super().__init__()

        self._type = "link"
        self.name = kwargs.get("name")
        self.materials = kwargs.get("materials", {})
        self.products = kwargs.get("products", {})
        self.byproducts = kwargs.get("byproducts", {})
        self.command = kwargs.get("command", [])
        self.environment = kwargs.get("environment", {})

        self.validate()

    @property
    def type_(self):
        """The string "link" to indentify the in-toto metadata type."""
        # NOTE: We expose the type_ property in the API documentation instead of
        # _type to protect it against modification.
        # NOTE: Trailing underscore is used by convention (pep8) to avoid conflict
        # with Python's type keyword.
        return self._type

    @staticmethod
    def read(data):
        """Creates a Link object from its dictionary representation.

        Arguments:
          data: A dictionary with link metadata fields.

        Raises:
          securesystemslib.exceptions.FormatError: Passed data is invalid.

        Returns:
          The created Link object.

        """
        return Link(**data)

    def _validate_type(self):
        """Private method to check that `_type` is set to "link"."""
        if self._type != "link":
            raise securesystemslib.exceptions.FormatError(
                "Invalid Link: field `_type` must be set to 'link', got: {}".format(
                    self._type
                )
            )

    def _validate_materials(self):
        """Private method to check that `materials` is a `dict` of `HASHDICTs`."""
        if not isinstance(self.materials, dict):
            raise securesystemslib.exceptions.FormatError(
                "Invalid Link: field `materials` must be of type dict, got: {}".format(
                    type(self.materials)
                )
            )

        for material in list(self.materials.values()):
            _check_hash_dict(material)

    def _validate_products(self):
        """Private method to check that `products` is a `dict` of `HASHDICTs`."""
        if not isinstance(self.products, dict):
            raise securesystemslib.exceptions.FormatError(
                "Invalid Link: field `products` must be of type dict, got: {}".format(
                    type(self.products)
                )
            )

        for product in list(self.products.values()):
            _check_hash_dict(product)

    def _validate_byproducts(self):
        """Private method to check that `byproducts` is a `dict`."""
        if not isinstance(self.byproducts, dict):
            raise securesystemslib.exceptions.FormatError(
                "Invalid Link: field `byproducts` must be of type dict, got: {}".format(
                    type(self.byproducts)
                )
            )

    def _validate_command(self):
        """Private method to check that `command` is a `list`."""
        if not isinstance(self.command, list):
            raise securesystemslib.exceptions.FormatError(
                "Invalid Link: field `command` must be of type list, got: {}".format(
                    type(self.command)
                )
            )

    def _validate_environment(self):
        """Private method to check that `environment` is a `dict`."""
        if not isinstance(self.environment, dict):
            raise securesystemslib.exceptions.FormatError(
                "Invalid Link: field `environment` must be of type dict, got: {}".format(
                    type(self.environment)
                )
            )



#!/usr/bin/env python

# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  in_toto_record.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  Nov 28, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides a command line interface for runlib.in_toto_record_start and
  runlib.in_toto_record_stop.

<Return Codes>
  2 if an exception occurred during argument parsing
  1 if an exception occurred
  0 if no exception occurred

"""
import argparse
import logging
import sys
from getpass import getpass

import in_toto.runlib
from in_toto import __version__
from in_toto.common_args import (
    BASE_PATH_ARGS,
    BASE_PATH_KWARGS,
    DSSE_ARGS,
    DSSE_KWARGS,
    EXCLUDE_ARGS,
    EXCLUDE_KWARGS,
    GPG_ARGS,
    GPG_HOME_ARGS,
    GPG_HOME_KWARGS,
    GPG_KWARGS,
    KEY_PASSWORD_ARGS,
    KEY_PASSWORD_KWARGS,
    LSTRIP_PATHS_ARGS,
    LSTRIP_PATHS_KWARGS,
    METADATA_DIRECTORY_ARGS,
    METADATA_DIRECTORY_KWARGS,
    OPTS_TITLE,
    QUIET_ARGS,
    QUIET_KWARGS,
    SIGNING_KEY_ARGS,
    SIGNING_KEY_KWARGS,
    VERBOSE_ARGS,
    VERBOSE_KWARGS,
    parse_password_and_prompt_args,
    sort_action_groups,
    title_case_action_groups,
)
from in_toto.models._signer import load_crypto_signer_from_pkcs8_file

# Command line interfaces should use in_toto base logger (c.f. in_toto.log)
LOG = logging.getLogger("in_toto")


def create_parser():
    """Create and return configured ArgumentParser instance."""

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
in-toto-record creates a signed link metadata file in two steps, in order to
provide evidence for supply chain steps that cannot be carried out by a single
command (for which 'in-toto-run' should be used). It returns a non-zero value
on failure and zero otherwise.""",
    )

    parser.epilog = """EXAMPLE USAGE

Create link metadata file in two commands, signing it with the private key
loaded from 'key_file', recording all files in the CWD as materials (on
start), and as products (on stop).

  {prog} start -n edit-files --signing-key path/to/key_file -m .
  {prog} stop -n edit-files --signing-key path/to/key_file -p .


Create link metadata file signed with the default GPG key from the default
GPG home directory and record a file named 'foo' as material and product.

  {prog} start -n edit-foo --gpg -m path/to/foo
  {prog} stop -n edit-foo --gpg -p path/to/foo


Create link metadata file signed with the private key loaded from 'key_file',
record all files in the CWD as material and product, and dump finished link
file to the target directory (on stop).

  {prog} start -n edit-files --signing-key path/to/key_file -m .
  {prog} stop -d path/to/target/dir -n edit-files --signing-key path/to/key_file -p .

""".format(
        prog=parser.prog
    )

    # The subparsers inherit the arguments from the parent parser
    parent_parser = argparse.ArgumentParser(add_help=False)
    subparsers = parser.add_subparsers(dest="command")

    # Workaround to make subcommands mandatory in Python>=3.3
    # https://bugs.python.org/issue9253#msg186387
    subparsers.required = True

    parent_named_args = parent_parser.add_argument_group(
        "required named arguments"
    )

    # FIXME: Do we limit the allowed characters for the name?
    parent_named_args.add_argument(
        "-n",
        "--step-name",
        type=str,
        required=True,
        metavar="<name>",
        help=(
            "name for the resulting link metadata file. It is also used to associate"
            " the link with a step defined in an in-toto layout."
        ),
    )

    parent_parser.add_argument(*KEY_PASSWORD_ARGS, **KEY_PASSWORD_KWARGS)

    parent_named_args.add_argument(*GPG_ARGS, **GPG_KWARGS)
    parent_parser.add_argument(*GPG_HOME_ARGS, **GPG_HOME_KWARGS)

    parent_named_args.add_argument(*SIGNING_KEY_ARGS, **SIGNING_KEY_KWARGS)

    parent_parser.add_argument(*EXCLUDE_ARGS, **EXCLUDE_KWARGS)
    parent_parser.add_argument(*BASE_PATH_ARGS, **BASE_PATH_KWARGS)
    parent_parser.add_argument(*LSTRIP_PATHS_ARGS, **LSTRIP_PATHS_KWARGS)
    parent_parser.add_argument(*DSSE_ARGS, **DSSE_KWARGS)

    verbosity_args = parent_parser.add_mutually_exclusive_group(required=False)
    verbosity_args.add_argument(*VERBOSE_ARGS, **VERBOSE_KWARGS)
    verbosity_args.add_argument(*QUIET_ARGS, **QUIET_KWARGS)

    subparser_start = subparsers.add_parser(
        "start",
        parents=[parent_parser],
        help=(
            "creates a preliminary link file recording the paths and hashes of"
            " the passed materials and signs it with the passed functionary's"
            " key. The resulting link file is stored as"
            " '.<name>.<keyid prefix>.link-unfinished'."
        ),
    )

    subparser_stop = subparsers.add_parser(
        "stop",
        parents=[parent_parser],
        help=(
            "expects preliminary link file '.<name>.<keyid prefix>.link-unfinished'"
            " in the CWD, signed by the passed functionary's key. If found, it"
            " records and adds the paths and hashes of the passed products to the"
            " link metadata file, updates the signature and renames the file to"
            " '<name>.<keyid prefix>.link'."
        ),
    )

    subparser_start.add_argument(
        "-m",
        "--materials",
        type=str,
        required=False,
        nargs="+",
        metavar="<path>",
        help=(
            "paths to files or directories, whose paths and hashes are stored in the"
            " resulting link metadata's material section when running the 'start'"
            " subcommand. Symlinks to files are followed."
        ),
    )

    subparser_stop.add_argument(
        "-p",
        "--products",
        type=str,
        required=False,
        nargs="+",
        metavar="<path>",
        help=(
            "paths to files or directories, whose paths and hashes are stored in the"
            " resulting link metadata's product section when running the 'stop'"
            " subcommand. Symlinks to files are followed."
        ),
    )

    subparser_stop.add_argument(
        *METADATA_DIRECTORY_ARGS, **METADATA_DIRECTORY_KWARGS
    )

    parser.add_argument(
        "--version",
        action="version",
        version="{} {}".format(parser.prog, __version__),
    )

    for parser_, order in [
        (parser, ["Positional Arguments", OPTS_TITLE]),
        (subparser_start, None),
        (subparser_stop, None),
    ]:
        title_case_action_groups(parser_)
        sort_action_groups(parser_, order)

    return parser


def main():
    """Parse arguments, load key from disk (if passed) and call
    either runlib.in_toto_record_start or runlib.in_toto_record_stop depending
    on the specified subcommand."""

    parser = create_parser()
    args = parser.parse_args()

    LOG.setLevelVerboseOrQuiet(args.verbose, args.quiet)

    # Use exactly one of gpg or pkcs8 signing key
    if sum([bool(args.gpg), bool(args.signing_key)]) != 1:
        parser.print_usage()
        parser.error(
            "Specify either '--signing-key <path>' or '--gpg [<keyid>]'"
        )

    password, prompt = parse_password_and_prompt_args(args)

    # If `--gpg` was set without argument it has the value `True` and
    # we will try to sign with the default key
    gpg_use_default = args.gpg is True

    # Otherwise gpg_keyid stays either None or gets the passed argument assigned
    gpg_keyid = None
    if not gpg_use_default and args.gpg:
        gpg_keyid = args.gpg

    try:
        signer = None
        if args.signing_key:
            if prompt:
                password = getpass()

            if password is not None:
                password = password.encode()

            signer = load_crypto_signer_from_pkcs8_file(
                args.signing_key, password
            )

        if args.command == "start":
            in_toto.runlib.in_toto_record_start(
                args.step_name,
                args.materials,
                gpg_keyid=gpg_keyid,
                gpg_use_default=gpg_use_default,
                gpg_home=args.gpg_home,
                exclude_patterns=args.exclude_patterns,
                base_path=args.base_path,
                lstrip_paths=args.lstrip_paths,
                use_dsse=args.use_dsse,
                signer=signer,
            )

        # Mutually exclusiveness is guaranteed by argparser
        else:  # args.command == "stop":
            in_toto.runlib.in_toto_record_stop(
                args.step_name,
                args.products,
                gpg_keyid=gpg_keyid,
                gpg_use_default=gpg_use_default,
                gpg_home=args.gpg_home,
                exclude_patterns=args.exclude_patterns,
                base_path=args.base_path,
                lstrip_paths=args.lstrip_paths,
                metadata_directory=args.metadata_directory,
                signer=signer,
            )

    except Exception as e:  # pylint: disable=broad-exception-caught
        LOG.error(
            "(in-toto-record %s) %s: %s", args.command, type(e).__name__, e
        )
        sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    main()



"""Resolver interface and implementations for files, OSTree, and directory
artifacts."""

import locale
import logging
import os
from abc import ABCMeta, abstractmethod
from functools import cmp_to_key
from itertools import combinations
from os.path import exists, isdir, isfile, join, normpath

from pathspec import GitIgnoreSpec
from securesystemslib.hash import digest, digest_filename

from in_toto.exceptions import PrefixError

logger = logging.getLogger(__name__)

_HASH_ALGORITHM = "sha256"

RESOLVER_FOR_URI_SCHEME = {}


class Resolver(metaclass=ABCMeta):
    """Resolver interface and factory."""

    @classmethod
    def for_uri(cls, uri):
        """Return registered resolver instance for passed URI."""
        scheme, match, _ = uri.partition(":")

        if not match or scheme not in RESOLVER_FOR_URI_SCHEME:
            scheme = FileResolver.SCHEME

        return RESOLVER_FOR_URI_SCHEME[scheme]

    @abstractmethod
    def hash_artifacts(self, uris):
        """Return hash dictionary for passed list of artifact URIs."""
        raise NotImplementedError


class FileResolver(Resolver):
    """File resolver implementation.

    Provides a ``hash_artifacts`` method to generate hashes for passed file
    paths. The resolver is configurable via its constructor.

    """

    SCHEME = "file"

    def __init__(
        self,
        exclude_patterns=None,
        base_path=None,
        follow_symlink_dirs=False,
        normalize_line_endings=False,
        lstrip_paths=None,
    ):
        if exclude_patterns is None:
            exclude_patterns = []

        if not lstrip_paths:
            lstrip_paths = []

        if base_path is not None:
            if not isinstance(base_path, str):
                raise ValueError("'base_path' must be string")

        for name, val in [
            ("exclude_patterns", exclude_patterns),
            ("lstrip_paths", lstrip_paths),
        ]:
            if not isinstance(val, list) or not all(
                isinstance(i, str) for i in val
            ):
                raise ValueError(f"'{name}' must be list of strings")

        for a_, b_ in combinations(lstrip_paths, 2):
            if a_.startswith(b_) or b_.startswith(a_):
                raise PrefixError(
                    f"'{a_}' and '{b_}' triggered a left substring error"
                )

        # Compile gitignore-style patterns
        self._exclude_filter = GitIgnoreSpec.from_lines(
            "gitwildmatch", exclude_patterns
        )
        self._base_path = base_path
        self._follow_symlink_dirs = follow_symlink_dirs
        self._normalize_line_endings = normalize_line_endings
        self._lstrip_paths = lstrip_paths

    def _exclude(self, path):
        """Helper to check, if path matches pre-compiled exclude patterns."""
        return self._exclude_filter.match_file(path)

    def _hash(self, path):
        """Helper to generate hash dictionary for path."""
        digest_obj = digest_filename(
            path,
            algorithm=_HASH_ALGORITHM,
            normalize_line_endings=self._normalize_line_endings,
        )
        return {_HASH_ALGORITHM: digest_obj.hexdigest()}

    def _mangle(self, path, existing_paths, scheme_prefix):
        """Helper for path mangling."""

        # Normalize slashes for cross-platform metadata consistency
        path = path.replace("\\", "/")

        # Left-strip names using configured path prefixes
        for lstrip_path in self._lstrip_paths:
            if path.startswith(lstrip_path):
                path = path[len(lstrip_path) :]
                break

        # Fail if left-stripping above results in duplicates
        if self._lstrip_paths and path in existing_paths:
            raise PrefixError(
                "Prefix selection has resulted in non unique dictionary key "
                f"'{path}'"
            )

        # Prepend passed scheme prefix
        path = scheme_prefix + path

        return path

    def _strip_scheme_prefix(self, path):
        """Helper to strip file resolver scheme prefix from path."""

        prefix = self.SCHEME + ":"
        if path.startswith(prefix):
            path = path[len(prefix) :]
        else:
            prefix = ""

        return path, prefix

    def hash_artifacts(self, uris):
        hashes = {}

        if self._base_path:
            original_cwd = os.getcwd()
            os.chdir(self._base_path)

        for path in uris:
            # Remove scheme prefix, but preserver to re-add later (see _mangle)
            path, prefix = self._strip_scheme_prefix(path)

            # Normalize URI before filtering and returning them
            # FIXME: Is this expected behavior? Does this make exclude patterns
            # with slashes platform-dependent? Check how 'gitwildmatch' treats
            # dots and slashes!
            path = normpath(path)

            if self._exclude(path):
                continue

            if not exists(path):
                logger.info("path: %s does not exist, skipping..", path)
                continue

            if isfile(path):
                name = self._mangle(path, hashes, prefix)
                hashes[name] = self._hash(path)

            if isdir(path):
                for base, dirs, names in os.walk(
                    path, followlinks=self._follow_symlink_dirs
                ):
                    # Filter directories to avoid unnecessary recursion below
                    # NOTE: Normalize to filter on directory paths without
                    # their dot-slash prefix, if path was dot.
                    dirs[:] = [
                        dirname
                        for dirname in dirs
                        if not self._exclude(normpath(join(base, dirname)))
                    ]

                    for filename in names:
                        # NOTE: Normalize to filter on and return file paths
                        # without their dot-slash prefix, if path was dot.
                        filepath = normpath(join(base, filename))

                        if self._exclude(filepath):
                            continue

                        if not isfile(filepath):
                            logger.info(
                                "File '%s' appears to be a broken symlink. "
                                "Skipping...",
                                filepath,
                            )
                            continue

                        name = self._mangle(filepath, hashes, prefix)
                        hashes[name] = self._hash(filepath)

        # Change back to original current working dir
        if self._base_path:
            os.chdir(original_cwd)

        return hashes


class OSTreeResolver(Resolver):
    """Resolver for OSTree repositories."""

    SCHEME = "ostree"

    # defining this separately as it should be pinned to OSTree's default
    # rather than in-toto's
    _HASH_ALGORITHM = "sha256"

    def __init__(self, base_path=None):
        self._base_path = base_path

    def _strip_scheme_prefix(self, path):
        """Helper to strip OSTree resolver scheme prefix from path."""

        return path[len(self.SCHEME + ":") :]

    def _add_scheme_prefix(self, path):
        """Helper to add scheme back after recording the hash."""

        return f"{self.SCHEME}:{path}"

    def _hash(self, path):
        """Helper to hash OSTree commits."""

        ref_path = os.path.join("refs", "heads", path)

        with open(ref_path, "r") as ref:  # pylint: disable=unspecified-encoding
            ref_contents = ref.read()
        ref_contents = ref_contents.strip("\n")

        object_path = os.path.join(
            "objects", ref_contents[:2], f"{ref_contents[2:]}.commit"
        )

        digest_obj = digest_filename(
            object_path,
            algorithm=self._HASH_ALGORITHM,
        )

        return {self._HASH_ALGORITHM: digest_obj.hexdigest()}

    def hash_artifacts(self, uris):
        hashes = {}

        if self._base_path:
            original_cwd = os.getcwd()
            os.chdir(self._base_path)

        for path in uris:
            # Remove scheme prefix, but preserver to re-add later
            path = self._strip_scheme_prefix(path)
            hashes[self._add_scheme_prefix(path)] = self._hash(path)

        # Change back to original current working dir
        if self._base_path:
            os.chdir(original_cwd)

        return hashes


class DirectoryResolver(Resolver):
    """Directory resolver implementation.

    Implementation of the following shell command for some directory URI:

    find . -type f | cut -c3- | LC_ALL=C sort | xargs -r sha256sum | sha256sum |
        cut -f1 -d' '

    `exclude_patterns` are not applied on the directory paths passed to
    `hash_artifacts`, but on the files inside the directory. Similarly,
    `follow_symlink_dirs` and `normalize_line_endings` are used on
    subdirectories and files inside the directories passed to `hash_artifacts`.

    """

    SCHEME = "dir"

    def __init__(
        self,
        exclude_patterns=None,
        follow_symlink_dirs=False,
        normalize_line_endings=False,
        lstrip_paths=None,
    ):
        if not exclude_patterns:
            exclude_patterns = []

        if not lstrip_paths:
            lstrip_paths = []

        self._exclude_patterns = exclude_patterns
        self._follow_symlink_dirs = follow_symlink_dirs
        self._normalize_line_endings = normalize_line_endings
        self._lstrip_paths = lstrip_paths

    def _strip_scheme_prefix(self, path):
        """Helper to strip file resolver scheme prefix from path."""

        return path[len(self.SCHEME + ":") :]

    def _mangle(self, path, existing_paths):
        """Helper for path mangling."""

        # Normalize slashes for cross-platform metadata consistency
        path = path.replace("\\", "/")

        # Left-strip names using configured path prefixes
        for lstrip_path in self._lstrip_paths:
            if path.startswith(lstrip_path):
                path = path[len(lstrip_path) :]
                break

        # Prepend passed scheme prefix
        # We do this first because the entries in existing_paths have the scheme
        # affixed already.
        path = self.SCHEME + ":" + path

        # Fail if left-stripping above results in duplicates
        if self._lstrip_paths and path in existing_paths:
            raise PrefixError(
                "Prefix selection has resulted in non unique dictionary key "
                f"'{path}'"
            )

        return path

    def _hash(self, file_hashes):
        """Helper to correctly sort and hash every element."""

        text_repr = ""
        keys = list(file_hashes.keys())
        if keys:
            locale.setlocale(locale.LC_ALL, "C")
            keys.sort(key=cmp_to_key(locale.strcoll))

            text_repr = "\n".join(
                [f"{file_hashes[k][_HASH_ALGORITHM]}  {k}" for k in keys]
            )
            text_repr += "\n"  # trailing new line

        digest_obj = digest(_HASH_ALGORITHM)
        digest_obj.update(text_repr.encode("utf-8"))

        return {_HASH_ALGORITHM: digest_obj.hexdigest()}

    def hash_artifacts(self, uris):
        hashes = {}

        for path in uris:
            path = self._strip_scheme_prefix(path)

            if not os.path.isdir(path):
                raise ValueError(f"path '{path}' is not a directory")

            file_resolver = FileResolver(
                base_path=path,
                exclude_patterns=self._exclude_patterns,
                follow_symlink_dirs=self._follow_symlink_dirs,
                normalize_line_endings=self._normalize_line_endings,
            )

            file_hashes = file_resolver.hash_artifacts(["."])
            if not file_hashes:
                logger.info(
                    "path: %s has no files, recording empty dir...", path
                )

            name = self._mangle(path, hashes)
            hashes[name] = self._hash(file_hashes)

        return hashes
